# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
import numpy as np
from typing import Union, TYPE_CHECKING, List, Dict
from PySide6 import QtWidgets
from PySide6.QtCore import Qt, QPoint, QPointF
from PySide6.QtGui import QPen, QCursor, QBrush, QColor
from PySide6.QtWidgets import QMenu, QGraphicsSceneMouseEvent

from VeraGridEngine.Devices.Fluid import FluidNode, FluidTurbine, FluidPump, FluidP2x
from VeraGridEngine.Devices.Substation.bus import Bus
from VeraGridEngine.enumerations import DeviceType, FaultType
from VeraGridEngine.Devices.Parents.editable_device import EditableDevice
from VeraGridEngine.Devices.types import FLUID_TYPES

from VeraGrid.Gui.Diagrams.generic_graphics import ACTIVE, FONT_SCALE, GenericDiagramWidget
from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import BarTerminalItem, HandleItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Fluid.fluid_turbine_graphics import FluidTurbineGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Fluid.fluid_pump_graphics import FluidPumpGraphicItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Fluid.fluid_p2x_graphics import FluidP2xGraphicItem
from VeraGrid.Gui.messages import yes_no_question, error_msg
from VeraGrid.Gui.gui_functions import add_menu_entry

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class FluidNodeGraphicItem(GenericDiagramWidget, QtWidgets.QGraphicsRectItem):
    """
      Represents a block in the diagram
      Has an x and y and width and height
      width and height can only be adjusted with a tip in the lower right corner.

      - in and output ports
      - parameters
      - description
    """

    def __init__(self, editor: SchematicWidget, fluid_node: FluidNode,
                 parent=None, index=0, h: float = 20, w: float = 80, x: float = 0, y: float = 0,
                 draw_labels: bool = True):

        GenericDiagramWidget.__init__(self, parent=parent, api_object=fluid_node, editor=editor,
                                      draw_labels=draw_labels)
        QtWidgets.QGraphicsRectItem.__init__(self, parent)

        self.min_w = 180.0
        self.min_h = 40.0
        self.offset = 20
        self.h = h if h >= self.min_h else self.min_h
        self.w = w if w >= self.min_w else self.min_w

        # loads, shunts, generators, etc...
        self._child_graphics = list()

        # Enabled for short circuit
        self.sc_enabled = [False, False, False, False]
        self.sc_type = FaultType.LLLG
        self.pen_width = 4

        # index
        self.index = index

        self.color = QColor(fluid_node.color) if fluid_node is not None else ACTIVE['fluid']
        self.style = ACTIVE['style']

        # Label:
        self.label = QtWidgets.QGraphicsTextItem(self.api_object.name if self.api_object is not None else "", self)
        self.label.setDefaultTextColor(ACTIVE['text'])
        self.label.setScale(FONT_SCALE)

        # square
        self.tile = QtWidgets.QGraphicsRectItem(0, 0, 20, 20, self)
        self.tile.setOpacity(0.7)

        # connection terminals the block
        self._terminal = BarTerminalItem('s', parent=self, editor=self.editor)  # , h=self.h))
        self._terminal.setPen(QPen(Qt.GlobalColor.transparent, self.pen_width, self.style,
                                   Qt.PenCapStyle.RoundCap, Qt.PenJoinStyle.RoundJoin))

        # Create corner for resize:
        self.sizer = HandleItem(self._terminal, callback=self.change_size)
        self.sizer.setPos(self.w, self.h)
        self.sizer.setFlag(self.GraphicsItemFlag.ItemIsMovable)

        self.big_marker = None

        self.set_tile_color(self.color)

        self.setPen(QPen(Qt.GlobalColor.transparent, self.pen_width, self.style))
        self.setBrush(Qt.GlobalColor.transparent)
        self.setFlags(self.GraphicsItemFlag.ItemIsSelectable | self.GraphicsItemFlag.ItemIsMovable)
        self.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

        # Update size:
        self.change_size(self.w, self.h)

        self.set_position(x, y)

    @property
    def api_object(self) -> FluidNode:
        return self._api_object

    @property
    def editor(self) -> SchematicWidget:
        return self._editor

    def get_associated_branch_graphics(self) -> List[GenericDiagramWidget]:
        """
        Get a list of all associated branch graphics
        :return:
        """
        conn: List[GenericDiagramWidget] = self._terminal.get_hosted_graphics()

        return conn

    def get_associated_widgets(self) -> List[GenericDiagramWidget]:
        """
        Get a list of all associated graphics
        :return:
        """
        conn: List[GenericDiagramWidget] = self.get_associated_branch_graphics()

        for graphics in self._child_graphics:
            conn.append(graphics)

        return conn

    def set_api_object_color(self):
        """
        Gather the color from the api object and apply
        :return:
        """
        self.color = QColor(self.api_object.color) if self.api_object is not None else ACTIVE['fluid']
        self.set_tile_color(self.color)

    def set_label(self, val: str):
        """
        Set the label content
        :param val:
        :return:
        """
        self.label.setPlainText(val)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """
        mouse press: display the editor
        :param event:
        :return:
        """
        if self.api_object is not None:
            self.editor.set_editor_model(api_object=self.api_object)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent):
        """
        On mouse move of this object...
        Args:
            event: QGraphicsSceneMouseEvent inherited
        """
        super().mouseMoveEvent(event)

        self.editor.update_diagram_element(device=self.api_object,
                                            x=self.pos().x(),
                                            y=self.pos().y(),
                                            w=self.w,
                                            h=self.h,
                                            r=self.rotation(),
                                            draw_labels=self.draw_labels,
                                            graphic_object=self)

    def set_position(self, x, y):
        """
        Set the bus x, y position
        :param x: x in pixels
        :param y: y in pixels
        """
        if np.isnan(x):
            x = 0
        if np.isnan(y):
            y = 0
        self.setPos(QPoint(int(x), int(y)))

    def recolour_mode(self):
        """
        Change the colour according to the system theme
        """
        # self.color = ACTIVE['color']
        # self.style = ACTIVE['style']
        self.label.setDefaultTextColor(ACTIVE['text'])
        # self.set_tile_color(self.color)

        for e in self._child_graphics:
            if e is not None:
                e.recolour_mode()

    def get_nexus_point(self) -> QPointF:
        """
        Get the connection point for the chldren nexus line
        :return: QPointF
        """
        return QPointF(self.x() + self.rect().width() / 2.0,
                       self.y() + self.rect().height() + self._terminal.h / 2.0)

    def set_tile_color(self, brush):
        """
        Set the color of the title
        Args:
            brush:  Qt Color
        """
        self.tile.setBrush(brush)
        self._terminal.setBrush(brush)

    def redraw(self):
        """
        Update the object
        :return:
        """
        self.change_size(self.w, self.h)

    def set_height(self, h):
        """
        Set the object height
        :param h: height in pixels
        """
        self.setRect(0.0, 0.0, self.w, h)
        self.h = h

    def change_size(self, w: float, h: Union[None, float] = None):
        """
        Resize block function
        @param w:
        @param h:
        @return:
        """
        # Limit the block size to the minimum size:
        self.w = w if w > self.min_w else self.min_w
        self.setRect(0.0, 0.0, self.w, self.min_h)
        y0 = self.offset
        x0 = 0

        # center label:
        self.label.setPos(self.w + 5, -20)

        # lower
        self._terminal.setPos(x0, y0)
        self._terminal.setRect(0, 20, self.w, 10)

        # rearrange children
        self.arrange_children()

        # update editor diagram position
        self.editor.update_diagram_element(device=self.api_object,
                                            x=self.pos().x(),
                                            y=self.pos().y(),
                                            w=self.w,
                                            h=int(self.min_h),
                                            r=self.rotation(),
                                            draw_labels=self.draw_labels,
                                            graphic_object=self)

        return self.w, self.min_h

    def arrange_children(self):
        """
        This function sorts the load and generators icons
        Returns:
            Nothing
        """
        y0 = self.h + 40
        n = len(self._child_graphics)
        inc_x = self.w / (n + 1)
        x = inc_x
        for elm in self._child_graphics:
            elm.setPos(x - elm.w / 2, y0)
            x += inc_x

        # Arrange line positions
        self._terminal.process_callbacks(self.pos() + self._terminal.pos())

    def create_children_widgets(self, injections_by_tpe: Dict[DeviceType, List[FLUID_TYPES]]):
        """
        Create the icons of the elements that are attached to the API bus object
        Returns:
            Nothing
        """
        for tpe, dev_list in injections_by_tpe.items():

            if tpe == DeviceType.FluidPumpDevice:
                for elm in dev_list:
                    self.add_pump(elm)

            elif tpe == DeviceType.FluidTurbineDevice:
                for elm in dev_list:
                    self.add_turbine(elm)

            elif tpe == DeviceType.FluidP2XDevice:
                for elm in dev_list:
                    self.add_p2x(elm)

            else:
                raise Exception("Unknown device type:" + str(tpe))

        self.arrange_children()

    def contextMenuEvent(self, event):
        """
        Display context menu
        @param event:
        @return:
        """
        menu = QMenu()
        menu.addSection("Fluid node")

        add_menu_entry(menu=menu,
                       text="Plot electrical profiles",
                       icon_path=":/Icons/icons/plot.png",
                       function_ptr=self.plot_electrical_profiles)

        add_menu_entry(menu=menu,
                       text="Plot fluid profiles",
                       icon_path=":/Icons/icons/plot.png",
                       function_ptr=self.plot_fluid_profiles)

        add_menu_entry(menu=menu,
                       text="Arrange",
                       icon_path=":/Icons/icons/automatic_layout.png",
                       function_ptr=self.arrange_children)

        add_menu_entry(menu=menu,
                       text="Delete all the connections",
                       icon_path=":/Icons/icons/delete_conn.png",
                       function_ptr=lambda: self.delete_all_connections(ask=True, delete_from_db=True))

        add_menu_entry(menu=menu,
                       text="Delete",
                       icon_path=":/Icons/icons/delete3.png",
                       function_ptr=self.remove)

        menu.addSection("Add")

        add_menu_entry(menu=menu,
                       text="Turbine",
                       icon_path=":/Icons/icons/add_gen.png",
                       function_ptr=self.add_turbine)

        add_menu_entry(menu=menu,
                       text="Pump",
                       icon_path=":/Icons/icons/add_gen.png",
                       function_ptr=self.add_pump)

        add_menu_entry(menu=menu,
                       text="P2X",
                       icon_path=":/Icons/icons/add_gen.png",
                       function_ptr=self.add_p2x)

        menu.exec_(event.screenPos())

    def get_terminal(self) -> BarTerminalItem:
        """
        Get the hosting terminal of this bus object
        :return: TerminalItem
        """
        return self._terminal

    def add_object(self, api_obj: Union[None, EditableDevice] = None):
        """
        Add any recognized object
        :param api_obj: EditableDevice
        """

        if api_obj.device_type == DeviceType.FluidTurbineDevice:
            self.add_turbine(api_obj=api_obj)

        elif api_obj.device_type == DeviceType.FluidPumpDevice:
            self.add_pump(api_obj=api_obj)

        elif api_obj.device_type == DeviceType.FluidP2XDevice:
            self.add_p2x(api_obj=api_obj)

        else:
            raise Exception("Cannot add device of type {}".format(api_obj.device_type.value))

    def create_bus_if_necessary(self) -> Bus:
        """
        Create the internal electrical bus of the fluid node
        :return api_object.bus
        """

        if self.api_object.bus is None:
            # create the bus and assign it
            self.api_object.bus = self.editor.circuit.add_bus()

            # set the same name
            self.api_object.bus.name = "Bus of " + self.api_object.name

        return self.api_object.bus

    def add_turbine(self, api_obj: Union[None, FluidTurbine] = None):
        """

        :param api_obj:
        :return:
        """
        if api_obj is None or type(api_obj) is bool:
            api_obj = self.editor.circuit.add_fluid_turbine(node=self.api_object, api_obj=None)
            api_obj.generator = self.editor.circuit.add_generator(bus=self.create_bus_if_necessary())
            api_obj.generator.name = f"Turbine @{self.api_object.name}"

        _grph = FluidTurbineGraphicItem(parent=self, api_obj=api_obj, editor=self.editor)
        self._child_graphics.append(_grph)
        self.arrange_children()
        return _grph

    def add_pump(self, api_obj: Union[None, FluidPump] = None):
        """

        :param api_obj:
        :return:
        """
        if api_obj is None or type(api_obj) is bool:
            api_obj = self.editor.circuit.add_fluid_pump(node=self.api_object, api_obj=None)
            api_obj.generator = self.editor.circuit.add_generator(bus=self.create_bus_if_necessary())
            api_obj.generator.name = f"Pump @{self.api_object.name}"

        _grph = FluidPumpGraphicItem(parent=self, api_obj=api_obj, editor=self.editor)
        self._child_graphics.append(_grph)
        self.arrange_children()
        return _grph

    def add_p2x(self, api_obj: Union[None, FluidP2x] = None):
        """

        :param api_obj:
        :return:
        """
        if api_obj is None or type(api_obj) is bool:
            api_obj = self.editor.circuit.add_fluid_p2x(node=self.api_object, api_obj=None)
            api_obj.generator = self.editor.circuit.add_generator(bus=self.create_bus_if_necessary())
            api_obj.generator.name = f"P2X @{self.api_object.name}"

        _grph = FluidP2xGraphicItem(parent=self, api_obj=api_obj, editor=self.editor)
        self._child_graphics.append(_grph)
        self.arrange_children()
        return _grph

    def delete_all_connections(self, ask: bool, delete_from_db: bool) -> None:
        """
        Delete all bus connections
        """
        if ask:
            ok = yes_no_question('Are you sure that you want to delete this fluid node',
                                 'Remove fluid node from schematic and DB' if delete_from_db else "Remove bus from schematic")
        else:
            ok = True

        if ok:
            self._terminal.remove_all_connections(delete_from_db=delete_from_db)

    def remove(self, ask=True):
        """
        Remove this element
        @return:
        """
        if ask:
            ok = yes_no_question('Are you sure that you want to delete this fluid node',
                                 'Remove fluid node')
        else:
            ok = True

        if ok:
            self.editor.remove_element(device=self.api_object, graphic_object=self)

    def update_color(self):
        """
        Update the colour
        """
        self.set_tile_color(QBrush(ACTIVE['color']))

    def plot_electrical_profiles(self):
        """

        @return:
        """
        # get the index of this object
        if self.api_object is not None:
            if self.api_object.bus is not None:
                i = self.editor.circuit.fluid_nodes.index(self.api_object)
                self.editor.plot_bus(i, self.api_object.bus)
            else:
                error_msg("No electrical bus attached :/")
        else:
            error_msg("No DB object attached :/")

    def plot_fluid_profiles(self):
        """

        @return:
        """
        # get the index of this object
        if self.api_object is not None:
            i = self.editor.circuit.fluid_nodes.index(self.api_object)
            self.editor.plot_fluid_node(i, self.api_object)
        else:
            error_msg("No DB object attached :/")

    def set_values(self, i: int, Vm: float, Va: float, P: float, Q: float,
                   tpe: str, format_str="{:10.2f}"):
        """

        :param i:
        :param Vm:
        :param Va:
        :param P:
        :param Q:
        :param tpe:
        :param format_str:
        :return:
        """
        if self.draw_labels:
            vm = format_str.format(Vm)

            if self.api_object is not None:
                if self.api_object.bus is not None:
                    vm_kv = format_str.format(Vm * self.api_object.bus.Vnom)
                else:
                    vm_kv = "No electrical bus"
            else:
                vm_kv = ""
            va = format_str.format(Va)
            msg = f"Bus {i}"
            if tpe is not None:
                msg += f" [{tpe}]"
            msg += "<br>"
            msg += f"v={vm}&lt;{va}º pu<br>"
            msg += f"V={vm_kv} kV<br>"
            if P is not None:
                p = format_str.format(P)
                q = format_str.format(Q)
                msg += f"P={p} MW<br>Q={q} MVAr"
        else:
            msg = ""

        title = self.api_object.name if self.api_object is not None else ""
        self.label.setHtml(f'<html><head/><body><p><span style=" font-size:10pt;">{title}<br/></span>'
                           f'<span style=" font-size:6pt;">{msg}</span></p></body></html>')

        self.setToolTip(msg)

    def set_fluid_values(self, i: int, Vm: float, Va: float, P: float, Q: float,
                         tpe: str,
                         fluid_node_p2x_flow: float,
                         fluid_node_current_level: float,
                         fluid_node_spillage: float,
                         fluid_node_flow_in: float,
                         fluid_node_flow_out: float,
                         format_str="{:10.2f}"):
        """

        :param i:
        :param Vm:
        :param Va:
        :param P:
        :param Q:
        :param tpe:
        :param fluid_node_p2x_flow:
        :param fluid_node_current_level:
        :param fluid_node_spillage:
        :param fluid_node_flow_in:
        :param fluid_node_flow_out:
        :param format_str:
        :return:
        """
        if self.draw_labels:
            vm = format_str.format(Vm)

            if self.api_object is not None:
                if self.api_object.bus is not None:
                    vm_kv = format_str.format(Vm * self.api_object.bus.Vnom)
                else:
                    vm_kv = "No electrical bus"
            else:
                vm_kv = ""
            va = format_str.format(Va)
            msg = f"Bus {i}"
            if tpe is not None:
                msg += f" [{tpe}]"
            msg += "<br>"
            msg += f"v={vm}&lt;{va}º pu<br>"
            msg += f"V={vm_kv} kV<br>"
            if P is not None:
                p = format_str.format(P)
                q = format_str.format(Q)
                msg += f"P={p} MW<br>Q={q} MVAr<br>"

            if fluid_node_flow_in is not None:
                f_in = format_str.format(fluid_node_flow_in)
                msg += f"In={f_in} m3/s<br>"

            if fluid_node_flow_out is not None:
                f_out = format_str.format(fluid_node_flow_out)
                msg += f"Out={f_out} m3/s<br>"

            if fluid_node_spillage is not None:
                f_spill = format_str.format(fluid_node_spillage)
                msg += f"Spill={f_spill} m3/s<br>"

            if fluid_node_current_level is not None:
                f_lvl = format_str.format(fluid_node_current_level)
                msg += f"Lvl={f_lvl} m3<br>"

            if fluid_node_p2x_flow is not None:
                f_p2x = format_str.format(fluid_node_p2x_flow)
                msg += f"P2X={f_p2x} m3/s<br>"

        else:
            msg = ""

        title = self.api_object.name if self.api_object is not None else ""
        self.label.setHtml(f'<html><head/><body><p><span style=" font-size:10pt;">{title}<br/></span>'
                           f'<span style=" font-size:6pt;">{msg}</span></p></body></html>')

        self.setToolTip(msg)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import os
import sys
import ctypes
import platform
import threading

PACKAGE_PARENT = '..'
SCRIPT_DIR = os.path.dirname(os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__))))
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))
from VeraGrid.__version__ import about_msg

# NOTE: For some reason I cannot begin to comprehend, the activation fails on windows if called before the GUI...
import VeraGridEngine.Utils.ThirdParty.gslv.gslv_activation


from VeraGrid.Gui.Main.VeraGridMain import runVeraGrid

# set the app name
APP_NAME = "veragrid"

if platform.system() == 'Windows' or os.name == 'nt':
    # this makes the icon display properly under windows
    myappid = u'mycompany.myproduct.subproduct.version'  # arbitrary string
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

elif os.name == "posix" or sys.platform.startswith("linux"):

    os.environ["QT_WAYLAND_CLIENT_IDENTIFIER"] = APP_NAME

    libc_name = ctypes.util.find_library("c")
    if libc_name:
        try:
            libc = ctypes.CDLL(libc_name)
            libc.prctl(15, APP_NAME.encode(), 0, 0, 0)
        except Exception as e:
            print(f"Could not set process name: {e}")


if __name__ == "__main__":
    print('Loading VeraGrid...')
    # check_all_svgs()
    print(about_msg)

    # Set the application style to the clear theme
    os.environ["QT_QUICK_CONTROLS_STYLE"] = "Basic"
    # os.environ["QT_QPA_PLATFORMTHEME"] = "qt5ct"  # this forces QT-only menus and look and feel
    threading.stack_size(134217728)
    runVeraGrid()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations
from typing import TYPE_CHECKING, Union
import numpy as np
from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon, QPixmap
from PySide6.QtWidgets import QMenu, QLabel, QDoubleSpinBox, QPushButton, QVBoxLayout, QComboBox, QDialog
from VeraGrid.Gui.gui_functions import get_list_model, add_menu_entry
from VeraGrid.Gui.messages import warning_msg
from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import BarTerminalItem, RoundTerminalItem
from VeraGridEngine.Devices.Branches.line import SequenceLineType, OverheadLineType, UndergroundLineType
from VeraGridEngine.Devices.Branches.dc_line import DcLine
from VeraGridEngine.enumerations import DeviceType
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.line_graphics_template import LineGraphicTemplateItem

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class DcLineEditor(QDialog):
    """
    DcLineEditor
    """
    def __init__(self, branch: DcLine, Sbase: float = 100.0, templates=None, current_template=None):
        """
        Line Editor constructor
        :param branch: Branch object to update
        :param Sbase: Base power in MVA
        """
        super(DcLineEditor, self).__init__()

        # keep pointer to the line object
        self.branch = branch

        self.Sbase = Sbase

        self.templates = templates

        self.current_template = current_template

        self.selected_template = None

        self.setObjectName("self")

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.NoContextMenu)

        self.layout = QVBoxLayout(self)

        # ------------------------------------------------------------------------------------------
        # Set the object values
        # ------------------------------------------------------------------------------------------
        Vf = self.branch.bus_from.Vnom
        Vt = self.branch.bus_to.Vnom

        Zbase = self.Sbase / (Vf * Vf)
        Ybase = 1 / Zbase

        R = self.branch.R * Zbase
        # X = self.branch.X * Zbase
        # B = self.branch.B * Ybase

        I = self.branch.rate / Vf  # current in kA

        # ------------------------------------------------------------------------------------------

        # catalogue
        self.catalogue_combo = QComboBox()
        if self.templates is not None:
            if len(self.templates) > 0:
                self.catalogue_combo.setModel(get_list_model(self.templates))

                if self.current_template is not None:
                    try:
                        idx = self.templates.index(self.current_template)
                        self.catalogue_combo.setCurrentIndex(idx)

                        if isinstance(self.current_template, SequenceLineType):
                            I = self.current_template.Imax
                            R = self.current_template.R

                        if isinstance(self.current_template, UndergroundLineType):
                            I = self.current_template.Imax
                            R = self.current_template.R

                        elif isinstance(self.current_template, OverheadLineType):
                            if self.current_template.check():
                                I = self.current_template.Imax[0]
                                R, X1, B1, I_kA = self.current_template.get_sequence_values(0, 1)
                            else:
                                warning_msg(text=f"The template {self.current_template.name} contains errors",
                                            title="Load template")

                    except:
                        pass

        # load template
        self.load_template_btn = QPushButton()
        self.load_template_btn.setText('Load template values')
        self.load_template_btn.clicked.connect(self.load_template_btn_click)

        # line length
        self.l_spinner = QDoubleSpinBox()
        self.l_spinner.setMinimum(0)
        self.l_spinner.setMaximum(9999999)
        self.l_spinner.setDecimals(6)
        self.l_spinner.setValue(1)

        # Max current
        self.i_spinner = QDoubleSpinBox()
        self.i_spinner.setMinimum(0)
        self.i_spinner.setMaximum(9999999)
        self.i_spinner.setDecimals(2)
        self.i_spinner.setValue(I)

        # R
        self.r_spinner = QDoubleSpinBox()
        self.r_spinner.setMinimum(0)
        self.r_spinner.setMaximum(9999999)
        self.r_spinner.setDecimals(6)
        self.r_spinner.setValue(R)

        # accept button
        self.accept_btn = QPushButton()
        self.accept_btn.setText('Accept')
        self.accept_btn.clicked.connect(self.accept_click)

        # add all to the GUI
        if templates is not None:
            self.layout.addWidget(QLabel("Available templates"))
            self.layout.addWidget(self.catalogue_combo)
            self.layout.addWidget(self.load_template_btn)
            self.layout.addWidget(QLabel(""))

        self.layout.addWidget(QLabel("L: Line length [km]"))
        self.layout.addWidget(self.l_spinner)

        self.layout.addWidget(QLabel("Imax: Max. current [kA] @" + str(int(Vf)) + " [kV]"))
        self.layout.addWidget(self.i_spinner)

        self.layout.addWidget(QLabel("R: Resistance [Ohm/km]"))
        self.layout.addWidget(self.r_spinner)

        # self.layout.addWidget(QLabel("X: Inductance [Ohm/km]"))
        # self.layout.addWidget(self.x_spinner)

        # self.layout.addWidget(QLabel("G: Conductance [S/km]"))
        # self.layout.addWidget(self.g_spinner)

        # self.layout.addWidget(QLabel("B: Susceptance [S/km]"))
        # self.layout.addWidget(self.b_spinner)

        self.layout.addWidget(self.accept_btn)

        self.setLayout(self.layout)

        self.setWindowTitle('Line editor')

    def accept_click(self):
        """
        Set the values
        :return:
        """
        l = self.l_spinner.value()
        I = self.i_spinner.value()
        R = self.r_spinner.value() * l

        Vf = self.branch.bus_from.Vnom
        Vt = self.branch.bus_to.Vnom

        Sn = np.round(I * Vf, 2)  # nominal power in MVA = kA * kV

        Zbase = self.Sbase / (Vf * Vf)
        Ybase = 1.0 / Zbase

        self.branch.R = np.round(R / Zbase, 6)
        self.branch.rate = Sn

        if self.selected_template is not None:
            self.branch.template = self.selected_template

        self.accept()

    def load_template(self, template):
        """

        :param template:
        :return:
        """
        if isinstance(template, SequenceLineType):
            self.i_spinner.setValue(template.Imax)
            self.r_spinner.setValue(template.R)

            self.selected_template = template

        elif isinstance(template, UndergroundLineType):
            self.i_spinner.setValue(template.Imax)
            self.r_spinner.setValue(template.R)

            self.selected_template = template

        elif isinstance(template, OverheadLineType):
            if self.current_template.check():
                R, X1, B1, I_kA = self.current_template.get_sequence_values(circuit_idx=0, seq=1)
                self.i_spinner.setValue(I_kA)
                self.r_spinner.setValue(R)

                self.selected_template = template
            else:
                warning_msg(text=f"The template {self.current_template.name} contains errors",
                            title="Load template")



    def load_template_btn_click(self):
        """
        Accept template values
        """

        if self.templates is not None:

            idx = self.catalogue_combo.currentIndex()
            template = self.templates[idx]

            self.load_template(template)


class DcLineGraphicItem(LineGraphicTemplateItem):

    def __init__(self,
                 from_port: Union[BarTerminalItem, RoundTerminalItem],
                 to_port: Union[BarTerminalItem, RoundTerminalItem],
                 editor: SchematicWidget,
                 width=5,
                 api_object: DcLine = None,
                 draw_labels: bool = True):
        """

        :param from_port:
        :param to_port:
        :param editor:
        :param width:
        :param api_object:
        """
        LineGraphicTemplateItem.__init__(self=self,
                                         from_port=from_port,
                                         to_port=to_port,
                                         editor=editor,
                                         width=width,
                                         api_object=api_object,
                                         draw_labels=draw_labels)

    @property
    def api_object(self) -> DcLine:
        return self._api_object

    def contextMenuEvent(self, event):
        """
        Show context menu
        @param event:
        @return:
        """
        if self.api_object is not None:
            menu = QMenu()
            menu.addSection("Line")

            pe = menu.addAction('Active')
            pe.setCheckable(True)
            pe.setChecked(self.api_object.active)
            pe.triggered.connect(self.enable_disable_toggle)

            add_menu_entry(menu=menu,
                           text="Draw labels",
                           icon_path="",
                           function_ptr=self.enable_disable_label_drawing,
                           checkeable=True,
                           checked_value=self.draw_labels)

            ra3 = menu.addAction('Editor')
            edit_icon = QIcon()
            edit_icon.addPixmap(QPixmap(":/Icons/icons/edit.png"))
            ra3.setIcon(edit_icon)
            ra3.triggered.connect(self.edit)

            rabf = menu.addAction('Change bus')
            move_bus_icon = QIcon()
            move_bus_icon.addPixmap(QPixmap(":/Icons/icons/move_bus.png"))
            rabf.setIcon(move_bus_icon)
            rabf.triggered.connect(self.change_bus)

            menu.addSeparator()

            ra6 = menu.addAction('Plot profiles')
            plot_icon = QIcon()
            plot_icon.addPixmap(QPixmap(":/Icons/icons/plot.png"))
            ra6.setIcon(plot_icon)
            ra6.triggered.connect(self.plot_profiles)

            ra4 = menu.addAction('Assign rate to profile')
            ra4_icon = QIcon()
            ra4_icon.addPixmap(QPixmap(":/Icons/icons/assign_to_profile.png"))
            ra4.setIcon(ra4_icon)
            ra4.triggered.connect(self.assign_rate_to_profile)

            ra5 = menu.addAction('Assign active state to profile')
            ra5_icon = QIcon()
            ra5_icon.addPixmap(QPixmap(":/Icons/icons/assign_to_profile.png"))
            ra5.setIcon(ra5_icon)
            ra5.triggered.connect(self.assign_status_to_profile)

            # menu.addSeparator()

            ra2 = menu.addAction('Delete')
            del_icon = QIcon()
            del_icon.addPixmap(QPixmap(":/Icons/icons/delete3.png"))
            ra2.setIcon(del_icon)
            ra2.triggered.connect(self.delete)

            menu.exec_(event.screenPos())
        else:
            pass

    def mouseDoubleClickEvent(self, event):
        """
        On double click, edit
        :param event:
        :return:
        """
        if self.api_object is not None:
            if self.api_object.device_type in [DeviceType.Transformer2WDevice, DeviceType.LineDevice]:
                # trigger the editor
                self.edit()
            elif self.api_object.device_type is DeviceType.SwitchDevice:
                # change state
                self.enable_disable_toggle()

    def edit(self):
        """
        Open the appropriate editor dialogue
        :return:
        """
        Sbase = self.editor.circuit.Sbase
        templates = self.editor.circuit.underground_cable_types + self.editor.circuit.overhead_line_types
        current_template = self.api_object.template
        dlg = DcLineEditor(self.api_object, Sbase, templates, current_template)
        if dlg.exec():
            pass

    def add_to_templates(self):
        """
        Open the appropriate editor dialogue
        :return:
        """
        Sbase = self.editor.circuit.Sbase

        dlg = DcLineEditor(branch=self.api_object, Sbase=Sbase)
        if dlg.exec():
            pass





# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
import time
import requests
import asyncio
from uuid import uuid4
from warnings import warn
from urllib3 import disable_warnings, exceptions
from typing import Callable, Dict, Union, List, Any
from PySide6.QtCore import QThread, Signal
from PySide6 import QtCore
from VeraGridEngine.basic_structures import Logger
from VeraGridEngine.Simulations.driver_handler import create_driver
from VeraGridEngine.IO.veragrid.remote import (gather_model_as_jsons_for_communication, RemoteInstruction, RemoteJob,
                                               send_json_data, get_certificate_path, get_certificate)
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.Simulations.types import DRIVER_OBJECTS

disable_warnings(exceptions.InsecureRequestWarning)


class JobsModel(QtCore.QAbstractTableModel):
    """
    Class to populate a Qt table view with a pandas data frame
    """

    def __init__(self) -> None:
        """
        """
        QtCore.QAbstractTableModel.__init__(self)
        self.jobs: List[RemoteJob] = list()
        self.headers = ["Job id", "User", "Grid name", "Job Type", "Status", "Progress"]

    def clear(self):
        """
        Clear jobs
        """
        self.jobs.clear()

    def parse_data(self, data: List[Dict[str, Union[str, Dict[str, Any]]]]):
        """
        Parse the data from the server
        :param data:
        :return:
        """
        self.jobs.clear()
        self.beginResetModel()
        for job_data in data:
            job = RemoteJob(data=job_data)
            self.jobs.append(job)

        self.endResetModel()

    def flags(self, index: QtCore.QModelIndex):
        """

        :param index:
        :return:
        """
        return QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable

    def rowCount(self, parent: Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex] = ...) -> int:
        """

        :param parent:
        :return:
        """
        return len(self.jobs)

    def columnCount(self, parent: Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex] = ...) -> int:
        """

        :param parent:
        :return:
        """
        return len(self.headers)

    def data(self, index: QtCore.QModelIndex, role: int = QtCore.Qt.ItemDataRole.DisplayRole) -> Any:
        """

        :param index:
        :param role:
        :return:
        """
        if index.isValid() and role == QtCore.Qt.ItemDataRole.DisplayRole:

            job = self.jobs[index.row()]

            # "id_tag", "Grid name", "Job Type", "Status"
            if index.column() == 0:
                return job.id_tag
            elif index.column() == 1:
                return job.instruction.user
            elif index.column() == 2:
                return job.grid_name
            elif index.column() == 3:
                return job.instruction.operation.value
            elif index.column() == 4:
                return job.status.value
            elif index.column() == 4:
                return job.progress
            else:
                return ""
        return None

    def setData(self, index, value, role=QtCore.Qt.ItemDataRole.DisplayRole):
        """

        :param index:
        :param value:
        :param role:
        :return:
        """
        return None

    def headerData(self,
                   section: int,
                   orientation: QtCore.Qt.Orientation,
                   role=QtCore.Qt.ItemDataRole.DisplayRole):
        """

        :param section:
        :param orientation:
        :param role:
        :return:
        """
        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if orientation == QtCore.Qt.Orientation.Horizontal:
                return self.headers[section]
            # elif orientation == QtCore.Qt.Orientation.Vertical:
            #     return self.jobs[section].id_tag

        return None


class ServerDriver(QThread):
    """
    Server driver
    """
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()
    connected_signal = Signal()
    sync_event = Signal()
    items_processed_event = Signal()

    def __init__(self, url: str, port: int, pwd: str, sleep_time: int = 2, status_func: Callable[[str], None] = None,
                 secure: bool = False):
        """
        Constructor
        :param url: Server URL
        :param port: Server port
        :param pwd: Server password
        :param sleep_time: Sleep time (s)
        :param status_func: a text function pointer
        """
        QThread.__init__(self)

        self.url = url
        self.port = port
        self.pwd = pwd
        self.sleep_time = sleep_time
        self.status_func: Callable[[str], None] = status_func
        self.secure = secure
        self.__running__ = False

        self.logger = Logger()

        self.data_model = JobsModel()

        self._loaded_certificate = False
        self._certificate_path = get_certificate_path()

        self.__cancel__ = False
        self.__pause__ = False

    def set_values(self, url: str, port: int, pwd: str, sleep_time: int = 2,
                   secure: bool = False,
                   status_func: Callable[[str], None] = None):
        """
        Set the values
        :param url: Server URL
        :param port: Server port
        :param pwd: Server password
        :param sleep_time: Sleep time (s)
        :param secure: Use https?
        :param status_func: a text function pointer
        """
        self.url = url
        self.port = port
        self.pwd = pwd
        self.sleep_time = sleep_time
        self.secure = secure
        self.status_func: Callable[[str], None] = status_func
        self.__running__ = False
        self.logger = Logger()

    def report_status(self, txt: str):
        """

        :param txt:
        :return:
        """
        if self.status_func is not None:
            self.status_func(txt)

    def base_url(self):
        """
        Base URL of the service
        :return:
        """
        if self.secure:
            return f"https://{self.url}:{self.port}"
        else:
            return f"http://{self.url}:{self.port}"

    def get_certificate_path(self):

        return self._certificate_path

    def is_running(self) -> bool:
        """
        Check if the server is running
        :return:
        """
        return self.__running__

    def get_server_certificate(self) -> bool:
        """
        Try get the server certificate
        :return: ok?
        """

        return get_certificate(base_url=self.base_url(),
                               certificate_path=self._certificate_path,
                               pwd=self.pwd,
                               logger=self.logger)

    def server_connect(self) -> bool:
        """
        Try connecting to the server
        :return: ok?
        """

        # get the SSL certificate (only once per class instance)
        self._loaded_certificate = self.get_server_certificate()
        self.__running__ = False

        # Make a GET request to the root endpoint
        try:
            response = requests.get(f"{self.base_url()}/",
                                    headers={"API-Key": self.pwd},
                                    verify=self._certificate_path,
                                    timeout=2)

            # Check if the request was successful
            if response.status_code == 200:
                # Print the response body
                # print("Response Body:", response.json())
                self.__running__ = True
                return True
            else:
                # Print error message
                self.logger.add_error(msg=f"Response error", value=response.text)
                print(response.text)
                return False
        except ConnectionError as e:
            self.logger.add_error(msg=f"Connection error", value=str(e))
            print(str(e))
            return False
        except Exception as e:
            self.logger.add_error(msg=f"General exception error", value=str(e))
            print(str(e))
            return False

    def get_jobs(self) -> bool:
        """
        Try connecting to the server
        :return: ok?
        """
        # Make a GET request to the root endpoint
        try:
            response = requests.get(f"{self.base_url()}/jobs_list",
                                    headers={"API-Key": self.pwd},
                                    verify=self._certificate_path,
                                    timeout=2)

            # Check if the request was successful
            if response.status_code == 200:
                # Parse the response body
                self.data_model.parse_data(data=response.json())
                return True
            else:
                # Print error message
                self.logger.add_error(msg=f"Response error", value=response.text)
                return False
        except ConnectionError as e:
            self.logger.add_error(msg=f"Connection error", value=str(e))
            return False
        except Exception as e:
            self.logger.add_error(msg=f"General exception error", value=str(e))
            return False

    def send_job(self, grid: MultiCircuit, instruction: RemoteInstruction) -> Dict[str, Any] | None:
        """
        
        :param grid:
        :param instruction:
        :return: 
        """
        websocket_url = f"{self.base_url()}/upload_job"

        if self.is_running():
            model_data = gather_model_as_jsons_for_communication(circuit=grid, instruction=instruction)

            response = asyncio.get_event_loop().run_until_complete(
                send_json_data(json_data=model_data,
                               endpoint_url=websocket_url,
                               certificate=self._certificate_path)
            )

            return response

        else:
            return None

    def delete_job(self, job_id: str, api_key: str) -> dict:
        """
        Delete a specific job by ID using the REST API.

        :param job_id: The ID of the job to delete_with_dialogue
        :param api_key: The API key for authentication
        :return: Response from the server
        """
        url = f"{self.base_url()}/jobs/{job_id}"
        headers = {
            "accept": "application/json",
            "API-Key": api_key
        }
        response = requests.delete(url, headers=headers, verify=self._certificate_path)

        if response.status_code == 200:
            return response.json()
        else:
            response.raise_for_status()

        self.get_jobs()

    def download_results(self, job_id: str, api_key: str, local_filename: str):
        """

        :param job_id:
        :param api_key:
        :param local_filename:
        :return:
        """
        url = f"{self.base_url()}/download_results/{job_id}"

        headers = {
            "accept": "application/json",
            "API-Key": api_key
        }

        print("Started download...")

        chunk_size = 1024 * 1024  # 1 MB
        sent = 0
        # Stream the download to avoid loading the entire file into memory
        with requests.get(url, headers=headers, stream=True, verify=self._certificate_path) as response:

            if response.status_code == 200:

                with open(local_filename, "wb") as file:
                    for chunk in response.iter_content(chunk_size=chunk_size):  # 1MB chunks
                        if chunk:  # Filter out keep-alive chunks
                            file.write(chunk)
                            sent += chunk_size
                            self.progress_text.emit(f"Sent {sent / chunk_size} MBytes")

            else:
                print(response.status_code, response.text)

        self.progress_text.emit(f"Downloaded file saved as {local_filename}")
        print(f"Downloaded file saved as {local_filename}")

    def cancel_job(self, job_id: str, api_key: str) -> dict:
        """
        Cancel a specific job by ID using the REST API.

        :param job_id: The ID of the job to cancel
        :param api_key: The API key for authentication
        :return: Response from the server
        """
        url = f"{self.base_url()}/jobs/{job_id}/cancel"
        headers = {
            "accept": "application/json",
            "API-Key": api_key
        }
        response = requests.post(url, headers=headers, verify=self._certificate_path)

        if response.status_code == 200:
            return response.json()
        else:
            response.raise_for_status()

        self.get_jobs()

    def run(self) -> None:
        """
        run the file save procedure
        """
        self.__cancel__ = False
        self.__pause__ = False

        self.report_status("Trying to connect")
        self._loaded_certificate = False  # set to false, so that we force re-download
        ok = self.server_connect()

        if ok:

            # get the running jobs
            self.get_jobs()

            self.report_status("Sync")
            self.connected_signal.emit()

        else:
            # bad connection
            self.report_status("Could not connect")
            self.__running__ = False
            self.done_signal.emit()
            return None

        # self.data_model.clear()
        # self.report_status("Sync stop")
        # self.__running__ = False
        # self.done_signal.emit()

    def cancel(self) -> None:
        """
        Cancel the sync checking
        """
        self.__running__ = False
        self.__cancel__ = True

    def pause(self):
        """
        Pause the sync checking
        """
        self.__pause__ = True

    def resume(self):
        """
        Resume the sync checking
        """
        self.__pause__ = False

    def process_issues(self):
        """
        Process all the issues
        :return:
        """

        self.items_processed_event.emit()


class RemoteJobDriver(QThread):
    """
    Server driver
    """
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal(str)
    sync_event = Signal()
    items_processed_event = Signal()

    def __init__(self,
                 grid: MultiCircuit,
                 instruction: RemoteInstruction,
                 base_url: str,
                 certificate_path: str,
                 register_driver_func) -> None:
        """

        :param grid:
        :param instruction:
        :param base_url:
        :param certificate_path:
        """
        QThread.__init__(self)

        self.idtag = uuid4().hex

        self.grid = grid
        self.instruction = instruction
        self.base_url = base_url
        self.certificate_path = certificate_path

        self.register_driver_func = register_driver_func

        self.logger = Logger()

    def run(self):
        """

        :return:
        """
        websocket_url = f"{self.base_url}/upload_job"

        model_data = gather_model_as_jsons_for_communication(circuit=self.grid, instruction=self.instruction)

        try:
            response, ok = send_json_data(json_data=model_data,
                                          endpoint_url=websocket_url,
                                          certificate=self.certificate_path)

            if not ok:
                response_text = response
                self.logger.add_error("Response error", value=response_text)
                response = None

        except requests.exceptions.ConnectionError as e:
            warn(str(e))
            response = None
            self.logger.add_error("Remote end closed connection without response")

        if response is not None:

            success = response.get("success", False)

            if success:
                results_data = response["results"]

                time_indices = results_data.get('time_indices', self.grid.get_all_time_indices())
                driver = create_driver(grid=self.grid,
                                       driver_tpe=self.instruction.operation,
                                       time_indices=time_indices)

                if driver is not None:
                    driver.results.parse_data(data=results_data)
                    self.register_driver_func(driver=driver)
            else:
                self.logger.add_error(msg=response.get("msg", "No message"))

        self.done_signal.emit(self.idtag)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'roseta_explorer.ui'
##
## Created by: Qt User Interface Compiler version 6.10.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QFrame, QGridLayout,
    QHBoxLayout, QHeaderView, QLabel, QLineEdit,
    QListView, QMainWindow, QMenu, QMenuBar,
    QProgressBar, QPushButton, QSizePolicy, QSpacerItem,
    QSplitter, QTabWidget, QTableView, QTreeView,
    QVBoxLayout, QWidget)
from VeraGrid.Gui.Icons.icons_rc import *

class Ui_RosetaExplorer(object):
    def setupUi(self, RosetaExplorer):
        if not RosetaExplorer.objectName():
            RosetaExplorer.setObjectName(u"RosetaExplorer")
        RosetaExplorer.resize(1267, 789)
        RosetaExplorer.setBaseSize(QSize(0, 0))
        RosetaExplorer.setAcceptDrops(True)
        icon = QIcon()
        icon.addFile(u":/Icons/icons/roseta.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        RosetaExplorer.setWindowIcon(icon)
        RosetaExplorer.setAutoFillBackground(False)
        RosetaExplorer.setIconSize(QSize(48, 48))
        RosetaExplorer.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)
        RosetaExplorer.setDocumentMode(False)
        RosetaExplorer.setTabShape(QTabWidget.TabShape.Rounded)
        RosetaExplorer.setDockNestingEnabled(False)
        RosetaExplorer.setUnifiedTitleAndToolBarOnMac(False)
        self.actionSave_logs = QAction(RosetaExplorer)
        self.actionSave_logs.setObjectName(u"actionSave_logs")
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/save.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionSave_logs.setIcon(icon1)
        self.centralwidget = QWidget(RosetaExplorer)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout_2 = QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.frame = QFrame(self.centralwidget)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.Shape.NoFrame)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout = QVBoxLayout(self.frame)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.mainTabWidget = QTabWidget(self.frame)
        self.mainTabWidget.setObjectName(u"mainTabWidget")
        self.dataTabLayout = QWidget()
        self.dataTabLayout.setObjectName(u"dataTabLayout")
        self.verticalLayout_4 = QVBoxLayout(self.dataTabLayout)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(6, 6, 6, 6)
        self.modelTabWidget = QTabWidget(self.dataTabLayout)
        self.modelTabWidget.setObjectName(u"modelTabWidget")
        self.modelTabWidget.setTabPosition(QTabWidget.TabPosition.South)
        self.tab_2 = QWidget()
        self.tab_2.setObjectName(u"tab_2")
        self.verticalLayout_5 = QVBoxLayout(self.tab_2)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.frame_5 = QFrame(self.tab_2)
        self.frame_5.setObjectName(u"frame_5")
        self.frame_5.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_5.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout_3 = QHBoxLayout(self.frame_5)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.treeFilterLineEdit = QLineEdit(self.frame_5)
        self.treeFilterLineEdit.setObjectName(u"treeFilterLineEdit")

        self.horizontalLayout_3.addWidget(self.treeFilterLineEdit)

        self.treeFilterComboBox = QComboBox(self.frame_5)
        self.treeFilterComboBox.setObjectName(u"treeFilterComboBox")

        self.horizontalLayout_3.addWidget(self.treeFilterComboBox)

        self.treeFilterButton = QPushButton(self.frame_5)
        self.treeFilterButton.setObjectName(u"treeFilterButton")

        self.horizontalLayout_3.addWidget(self.treeFilterButton)


        self.verticalLayout_5.addWidget(self.frame_5)

        self.mainTreeView = QTreeView(self.tab_2)
        self.mainTreeView.setObjectName(u"mainTreeView")

        self.verticalLayout_5.addWidget(self.mainTreeView)

        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/array.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.modelTabWidget.addTab(self.tab_2, icon2, "")
        self.modelTab = QWidget()
        self.modelTab.setObjectName(u"modelTab")
        self.verticalLayout_10 = QVBoxLayout(self.modelTab)
        self.verticalLayout_10.setObjectName(u"verticalLayout_10")
        self.main_splitter = QSplitter(self.modelTab)
        self.main_splitter.setObjectName(u"main_splitter")
        self.main_splitter.setOrientation(Qt.Orientation.Horizontal)
        self.frame_2 = QFrame(self.main_splitter)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_6 = QVBoxLayout(self.frame_2)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.frame_7 = QFrame(self.frame_2)
        self.frame_7.setObjectName(u"frame_7")
        self.frame_7.setMinimumSize(QSize(0, 26))
        self.frame_7.setMaximumSize(QSize(16777215, 16777215))
        self.frame_7.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_7.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout_2 = QHBoxLayout(self.frame_7)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.modelTypeLabel = QLabel(self.frame_7)
        self.modelTypeLabel.setObjectName(u"modelTypeLabel")

        self.horizontalLayout_2.addWidget(self.modelTypeLabel)

        self.horizontalSpacer = QSpacerItem(529, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer)


        self.verticalLayout_6.addWidget(self.frame_7)

        self.clasesListView = QListView(self.frame_2)
        self.clasesListView.setObjectName(u"clasesListView")

        self.verticalLayout_6.addWidget(self.clasesListView)

        self.main_splitter.addWidget(self.frame_2)
        self.frame_6 = QFrame(self.main_splitter)
        self.frame_6.setObjectName(u"frame_6")
        self.frame_6.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_6.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_3 = QVBoxLayout(self.frame_6)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(1, 1, 1, 1)
        self.frame_4 = QFrame(self.frame_6)
        self.frame_4.setObjectName(u"frame_4")
        self.frame_4.setMaximumSize(QSize(16777215, 40))
        self.frame_4.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_4.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout = QHBoxLayout(self.frame_4)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.filterLineEdit = QLineEdit(self.frame_4)
        self.filterLineEdit.setObjectName(u"filterLineEdit")

        self.horizontalLayout.addWidget(self.filterLineEdit)

        self.filterComboBox = QComboBox(self.frame_4)
        self.filterComboBox.setObjectName(u"filterComboBox")

        self.horizontalLayout.addWidget(self.filterComboBox)

        self.filterButton = QPushButton(self.frame_4)
        self.filterButton.setObjectName(u"filterButton")

        self.horizontalLayout.addWidget(self.filterButton)


        self.verticalLayout_3.addWidget(self.frame_4)

        self.propertiesTableView = QTableView(self.frame_6)
        self.propertiesTableView.setObjectName(u"propertiesTableView")

        self.verticalLayout_3.addWidget(self.propertiesTableView)

        self.main_splitter.addWidget(self.frame_6)

        self.verticalLayout_10.addWidget(self.main_splitter)

        self.modelTabWidget.addTab(self.modelTab, icon2, "")
        self.loggerTab = QWidget()
        self.loggerTab.setObjectName(u"loggerTab")
        self.verticalLayout_9 = QVBoxLayout(self.loggerTab)
        self.verticalLayout_9.setObjectName(u"verticalLayout_9")
        self.loggerTreeView = QTreeView(self.loggerTab)
        self.loggerTreeView.setObjectName(u"loggerTreeView")

        self.verticalLayout_9.addWidget(self.loggerTreeView)

        icon3 = QIcon()
        icon3.addFile(u":/Icons/icons/data.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.modelTabWidget.addTab(self.loggerTab, icon3, "")

        self.verticalLayout_4.addWidget(self.modelTabWidget)

        self.mainTabWidget.addTab(self.dataTabLayout, icon2, "")
        self.tab = QWidget()
        self.tab.setObjectName(u"tab")
        self.horizontalLayout_4 = QHBoxLayout(self.tab)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.consoleLayout = QVBoxLayout()
        self.consoleLayout.setObjectName(u"consoleLayout")

        self.horizontalLayout_4.addLayout(self.consoleLayout)

        icon4 = QIcon()
        icon4.addFile(u":/Icons/icons/console.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.mainTabWidget.addTab(self.tab, icon4, "")

        self.verticalLayout.addWidget(self.mainTabWidget)


        self.verticalLayout_2.addWidget(self.frame)

        self.progress_frame = QFrame(self.centralwidget)
        self.progress_frame.setObjectName(u"progress_frame")
        self.progress_frame.setFrameShape(QFrame.Shape.NoFrame)
        self.progress_frame.setFrameShadow(QFrame.Shadow.Raised)
        self.gridLayout_7 = QGridLayout(self.progress_frame)
        self.gridLayout_7.setObjectName(u"gridLayout_7")
        self.cancelButton = QPushButton(self.progress_frame)
        self.cancelButton.setObjectName(u"cancelButton")
        self.cancelButton.setMinimumSize(QSize(0, 24))
        icon5 = QIcon()
        icon5.addFile(u":/Icons/icons/delete.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.cancelButton.setIcon(icon5)

        self.gridLayout_7.addWidget(self.cancelButton, 1, 0, 1, 1)

        self.progressBar = QProgressBar(self.progress_frame)
        self.progressBar.setObjectName(u"progressBar")
        palette = QPalette()
        brush = QBrush(QColor(159, 159, 159, 255))
        brush.setStyle(Qt.BrushStyle.SolidPattern)
        palette.setBrush(QPalette.ColorGroup.Active, QPalette.ColorRole.Text, brush)
        brush1 = QBrush(QColor(159, 159, 159, 128))
        brush1.setStyle(Qt.BrushStyle.SolidPattern)
#if QT_VERSION >= QT_VERSION_CHECK(5, 12, 0)
        palette.setBrush(QPalette.ColorGroup.Active, QPalette.ColorRole.PlaceholderText, brush1)
#endif
        palette.setBrush(QPalette.ColorGroup.Inactive, QPalette.ColorRole.Text, brush)
#if QT_VERSION >= QT_VERSION_CHECK(5, 12, 0)
        palette.setBrush(QPalette.ColorGroup.Inactive, QPalette.ColorRole.PlaceholderText, brush1)
#endif
        brush2 = QBrush(QColor(120, 120, 120, 255))
        brush2.setStyle(Qt.BrushStyle.SolidPattern)
        palette.setBrush(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text, brush2)
        brush3 = QBrush(QColor(0, 0, 0, 128))
        brush3.setStyle(Qt.BrushStyle.SolidPattern)
#if QT_VERSION >= QT_VERSION_CHECK(5, 12, 0)
        palette.setBrush(QPalette.ColorGroup.Disabled, QPalette.ColorRole.PlaceholderText, brush3)
#endif
        self.progressBar.setPalette(palette)
        self.progressBar.setAutoFillBackground(False)
        self.progressBar.setStyleSheet(u"QProgressBar {\n"
"	border: 1px solid rgb(186, 189, 182);\n"
"    border-radius: 5px;\n"
"	text-align: center;\n"
"}\n"
"QProgressBar::chunk{\n"
"	background-color: rgb(0, 34, 43);\n"
"    color: rgb(255, 255, 255)\n"
"}\n"
"")
        self.progressBar.setValue(50)
        self.progressBar.setTextVisible(True)
        self.progressBar.setInvertedAppearance(False)

        self.gridLayout_7.addWidget(self.progressBar, 1, 1, 1, 1)

        self.progressLabel = QLabel(self.progress_frame)
        self.progressLabel.setObjectName(u"progressLabel")

        self.gridLayout_7.addWidget(self.progressLabel, 0, 1, 1, 1)


        self.verticalLayout_2.addWidget(self.progress_frame)

        RosetaExplorer.setCentralWidget(self.centralwidget)
        self.menuBar = QMenuBar(RosetaExplorer)
        self.menuBar.setObjectName(u"menuBar")
        self.menuBar.setGeometry(QRect(0, 0, 1267, 23))
        self.menuFile = QMenu(self.menuBar)
        self.menuFile.setObjectName(u"menuFile")
        RosetaExplorer.setMenuBar(self.menuBar)

        self.menuBar.addAction(self.menuFile.menuAction())
        self.menuFile.addAction(self.actionSave_logs)

        self.retranslateUi(RosetaExplorer)

        self.mainTabWidget.setCurrentIndex(0)
        self.modelTabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(RosetaExplorer)
    # setupUi

    def retranslateUi(self, RosetaExplorer):
        RosetaExplorer.setWindowTitle(QCoreApplication.translate("RosetaExplorer", u"Cgmes Explorer", None))
        self.actionSave_logs.setText(QCoreApplication.translate("RosetaExplorer", u"Save logs", None))
        self.treeFilterButton.setText(QCoreApplication.translate("RosetaExplorer", u"Filter", None))
        self.modelTabWidget.setTabText(self.modelTabWidget.indexOf(self.tab_2), QCoreApplication.translate("RosetaExplorer", u"Tree view", None))
        self.modelTypeLabel.setText(QCoreApplication.translate("RosetaExplorer", u"...", None))
        self.filterButton.setText(QCoreApplication.translate("RosetaExplorer", u"Filter", None))
        self.modelTabWidget.setTabText(self.modelTabWidget.indexOf(self.modelTab), QCoreApplication.translate("RosetaExplorer", u"Table view", None))
        self.modelTabWidget.setTabText(self.modelTabWidget.indexOf(self.loggerTab), QCoreApplication.translate("RosetaExplorer", u"Logger", None))
        self.mainTabWidget.setTabText(self.mainTabWidget.indexOf(self.dataTabLayout), QCoreApplication.translate("RosetaExplorer", u"Data", None))
        self.mainTabWidget.setTabText(self.mainTabWidget.indexOf(self.tab), QCoreApplication.translate("RosetaExplorer", u"Python", None))
#if QT_CONFIG(tooltip)
        self.cancelButton.setToolTip(QCoreApplication.translate("RosetaExplorer", u"Cancel process", None))
#endif // QT_CONFIG(tooltip)
        self.cancelButton.setText("")
        self.progressLabel.setText("")
        self.menuFile.setTitle(QCoreApplication.translate("RosetaExplorer", u"File", None))
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from typing import Dict, List
import numpy as np
from matplotlib import pyplot as plt
from PySide6 import QtCore, QtGui

from VeraGrid.Gui.Icons.icon_associations import device_type_icons
from VeraGridEngine.Devices.types import ALL_DEV_TYPES
from VeraGridEngine.Simulations.Rms.rms_results import RmsResults
from VeraGridEngine.Utils.Symbolic.symbolic import Var
from VeraGridEngine.enumerations import DeviceType
from VeraGridEngine.Simulations.results_table import ResultsTable
from VeraGrid.Gui.results_model import ResultsModel


def _build_tree_item(text: str) -> QtGui.QStandardItem:
    """
    Build a non-editable tree item.

    :param text: Text to display in the tree node.
    :return: Configured tree item.
    """
    # Every tree node is marked as read-only because the trees are used as selectors and drop targets.
    item: QtGui.QStandardItem = QtGui.QStandardItem(text)
    item.setEditable(False)
    return item


def _get_device_type_label(device_tpe: DeviceType) -> str:
    """
    Get the label that represents a device type.

    :param device_tpe: Device type enum value.
    :return: User-facing label for the device type node.
    """
    # The enum value is already the canonical label used across the GUI.
    return str(device_tpe.value)


def _get_device_label(device: ALL_DEV_TYPES) -> str:
    """
    Get the label that represents a device instance.

    :param device: Device instance stored in the RMS results tree.
    :return: User-facing label for the device node.
    """
    # Engine devices expose the correct GUI name through their string representation.
    return str(device)


def _get_var_label(variable: Var) -> str:
    """
    Get the label that represents a simulation variable.

    :param variable: RMS variable object.
    :return: User-facing label for the variable node.
    """
    # The symbolic variable name is the exact identifier used by the simulation arrays.
    return variable.name


def _set_item_icon(item: QtGui.QStandardItem, icon_key: str) -> None:
    """
    Set the item icon from the shared device-type icon dictionary.

    :param item: Tree item that should receive the icon.
    :param icon_key: Key used to search in ``device_type_icons``.
    :return: Nothing.
    """
    # The icon mapping is shared with the rest of the GUI so tree categories stay visually consistent.
    icon_path: str | None = device_type_icons.get(icon_key, None)
    if icon_path is not None:
        icon: QtGui.QIcon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(icon_path))
        item.setIcon(icon)
    else:
        pass


class DynamicsPlotGroup:
    """
    Group of variables to be plotted together.
    """

    __slots__ = ("_name", "_vars")

    def __init__(self, name: str):
        """
        Build the plot group.

        :param name: Name shown in the plots tree.
        """
        # The name is the external identifier used by the UI and CRUD operations.
        self._name: str = name

        # The variable list is kept ordered because users expect insertion order to be preserved in the plot tree.
        self._vars: List[Var] = list()

    def get_name(self) -> str:
        """
        Get the plot-group name.

        :return: Plot-group name.
        """
        return self._name

    def set_name(self, name: str) -> None:
        """
        Set the plot-group name.

        :param name: New group name.
        :return: Nothing.
        """
        self._name = name

    def get_vars(self) -> List[Var]:
        """
        Get the variables stored in the group.

        :return: Variables in insertion order.
        """
        return list(self._vars)

    def contains_var(self, variable: Var) -> bool:
        """
        Check whether a variable already belongs to the group.

        :param variable: Variable to inspect.
        :return: ``True`` when the variable is already present.
        """
        contained: bool = False
        existing_var: Var
        for existing_var in self._vars:
            if existing_var.uid == variable.uid:
                contained = True
            else:
                pass
        return contained

    def add_var(self, variable: Var) -> bool:
        """
        Add a variable to the group.

        :param variable: Variable to add.
        :return: ``True`` when the variable was inserted.
        """
        # Duplicate insertions are rejected so drag-and-drop remains idempotent.
        if self.contains_var(variable=variable):
            return False
        else:
            self._vars.append(variable)
            return True

    def remove_var(self, variable: Var) -> bool:
        """
        Remove a variable from the group.

        :param variable: Variable to remove.
        :return: ``True`` when the variable was present and removed.
        """
        variable_idx: int = -1
        idx: int
        existing_var: Var
        for idx, existing_var in enumerate(self._vars):
            if existing_var.uid == variable.uid:
                variable_idx = idx
            else:
                pass

        if variable_idx >= 0:
            del self._vars[variable_idx]
            return True
        else:
            return False


class DynamicsPlotGroups:
    """
    Collection of plot groups with explicit CRUD operations.
    """

    __slots__ = ("_groups",)

    def __init__(self) -> None:
        """
        Build an empty plot-group collection.

        :return: Nothing.
        """
        self._groups: List[DynamicsPlotGroup] = list()

    def get_groups(self) -> List[DynamicsPlotGroup]:
        """
        Get the stored plot groups.

        :return: Copy of the plot-group list.
        """
        return list(self._groups)

    def get_group(self, name: str) -> DynamicsPlotGroup | None:
        """
        Get one plot group by name.

        :param name: Plot-group name.
        :return: Matching plot group or ``None``.
        """
        group: DynamicsPlotGroup
        for group in self._groups:
            if group.get_name() == name:
                return group
            else:
                pass
        return None

    def create_group(self, name: str) -> bool:
        """
        Create a new plot group.

        :param name: Requested group name.
        :return: ``True`` when the group was created.
        """
        clean_name: str = name.strip()
        if clean_name == "":
            return False
        else:
            existing_group: DynamicsPlotGroup | None = self.get_group(name=clean_name)
            if existing_group is None:
                self._groups.append(DynamicsPlotGroup(name=clean_name))
                return True
            else:
                return False

    def delete_group(self, name: str) -> bool:
        """
        Delete a plot group.

        :param name: Group name.
        :return: ``True`` when the group existed and was deleted.
        """
        group_idx: int = -1
        idx: int
        group: DynamicsPlotGroup
        for idx, group in enumerate(self._groups):
            if group.get_name() == name:
                group_idx = idx
            else:
                pass

        if group_idx >= 0:
            del self._groups[group_idx]
            return True
        else:
            return False

    def rename_group(self, old_name: str, new_name: str) -> bool:
        """
        Rename an existing group.

        :param old_name: Current group name.
        :param new_name: Requested group name.
        :return: ``True`` when the rename was applied.
        """
        clean_name: str = new_name.strip()
        if clean_name == "":
            return False
        else:
            target_group: DynamicsPlotGroup | None = self.get_group(name=old_name)
            clashing_group: DynamicsPlotGroup | None = self.get_group(name=clean_name)
            if target_group is not None:
                if clashing_group is None or clean_name == old_name:
                    target_group.set_name(name=clean_name)
                    return True
                else:
                    return False
            else:
                return False


class DynamicsDeviceTreeModel(QtGui.QStandardItemModel):
    """
    Source tree model that exports dragged variables.
    """

    __slots__ = ("_var_role", "_mime_type")

    def __init__(self, var_role: int, mime_type: str):
        """
        Build the source dynamics tree model.

        :param var_role: Qt role that stores the ``Var`` object in leaf nodes.
        :param mime_type: Mime type used during drag-and-drop.
        """
        QtGui.QStandardItemModel.__init__(self)
        self._var_role: int = var_role
        self._mime_type: str = mime_type

    def flags(self, index: QtCore.QModelIndex) -> QtCore.Qt.ItemFlag:
        """
        Get the Qt flags for an item.

        :param index: Model index.
        :return: Flags controlling selection and dragging.
        """
        if index.isValid():
            item: QtGui.QStandardItem | None = self.itemFromIndex(index)
            if item is not None:
                item_data: object = item.data(self._var_role)
                if isinstance(item_data, Var):
                    return (QtCore.Qt.ItemFlag.ItemIsEnabled
                            | QtCore.Qt.ItemFlag.ItemIsSelectable
                            | QtCore.Qt.ItemFlag.ItemIsDragEnabled)
                else:
                    return QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable
            else:
                return QtCore.Qt.ItemFlag.ItemIsEnabled
        else:
            return QtCore.Qt.ItemFlag.ItemIsEnabled

    def mimeTypes(self) -> List[str]:
        """
        Get the supported mime types.

        :return: Mime types exported by the model.
        """
        return [self._mime_type]

    def supportedDragActions(self) -> QtCore.Qt.DropAction:
        """
        Get the supported drag actions.

        :return: Copy action.
        """
        return QtCore.Qt.DropAction.CopyAction

    def mimeData(self, indexes: List[QtCore.QModelIndex]) -> QtCore.QMimeData:
        """
        Build mime data from the selected variable item.

        :param indexes: Selected indexes.
        :return: Mime payload containing the dragged variable uid.
        """
        mime_data: QtCore.QMimeData = QtCore.QMimeData()

        index: QtCore.QModelIndex
        for index in indexes:
            if index.isValid():
                item: QtGui.QStandardItem | None = self.itemFromIndex(index)
                if item is not None:
                    item_data: object = item.data(self._var_role)
                    if isinstance(item_data, Var):
                        mime_data.setData(self._mime_type, QtCore.QByteArray(str(item_data.uid).encode("utf-8")))
                        return mime_data
                    else:
                        pass
                else:
                    pass
            else:
                pass

        return mime_data


class DynamicsPlotsTreeModel(QtGui.QStandardItemModel):
    """
    Target tree model that accepts dropped variables into plot groups.
    """

    __slots__ = ("_handler",)

    def __init__(self, handler: "DynamicsResultsHandler"):
        """
        Build the plots tree model.

        :param handler: Dynamics-results handler that owns the plot-group state.
        """
        QtGui.QStandardItemModel.__init__(self)
        self._handler: DynamicsResultsHandler = handler

    def flags(self, index: QtCore.QModelIndex) -> QtCore.Qt.ItemFlag:
        """
        Get the Qt flags for a plots-tree item.

        :param index: Model index.
        :return: Flags controlling selection and drop acceptance.
        """
        if index.isValid():
            item: QtGui.QStandardItem | None = self.itemFromIndex(index)
            if item is not None:
                group_name_data: object = item.data(self._handler.get_group_name_role())
                if isinstance(group_name_data, str):
                    return (QtCore.Qt.ItemFlag.ItemIsEnabled
                            | QtCore.Qt.ItemFlag.ItemIsSelectable
                            | QtCore.Qt.ItemFlag.ItemIsDropEnabled)
                else:
                    return QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable
            else:
                return QtCore.Qt.ItemFlag.ItemIsEnabled
        else:
            return QtCore.Qt.ItemFlag.ItemIsEnabled

    def mimeTypes(self) -> List[str]:
        """
        Get the supported mime types.

        :return: Mime type accepted by the target model.
        """
        return [self._handler.get_drag_mime_type()]

    def supportedDropActions(self) -> QtCore.Qt.DropAction:
        """
        Get the supported drop actions.

        :return: Copy action.
        """
        return QtCore.Qt.DropAction.CopyAction

    def dropMimeData(self,
                     data: QtCore.QMimeData,
                     action: QtCore.Qt.DropAction,
                     row: int,
                     column: int,
                     parent: QtCore.QModelIndex) -> bool:
        """
        Handle dropping one variable into a plot group.

        :param data: Drag payload.
        :param action: Requested drop action.
        :param row: Drop row.
        :param column: Drop column.
        :param parent: Drop parent index.
        :return: ``True`` when the drop was accepted.
        """
        del row
        del column

        if action == QtCore.Qt.DropAction.IgnoreAction:
            return True
        else:
            if data.hasFormat(self._handler.get_drag_mime_type()):
                payload: bytes = bytes(data.data(self._handler.get_drag_mime_type()))
                payload_text: str = payload.decode("utf-8").strip()
                if payload_text.isdigit():
                    var_uid: int = int(payload_text)
                    group_name: str | None = self._handler.get_group_name_from_drop_index(index=parent)
                    if group_name is not None:
                        return self._handler.add_var_to_group(group_name=group_name, var_uid=var_uid)
                    else:
                        return False
                else:
                    return False
            else:
                return False


def build_dynamics_tree_model(tree_data: Dict[DeviceType, Dict[ALL_DEV_TYPES, List[Var]]],
                              var_role: int,
                              mime_type: str) -> DynamicsDeviceTreeModel:
    """
    Build the source tree-view model for RMS dynamics results.

    :param tree_data: Hierarchical RMS results tree grouped by device type and device.
    :param var_role: Qt item-data role used to store the ``Var`` instance in leaf nodes.
    :param mime_type: Mime type exported when dragging a variable.
    :return: Source tree model ready to be assigned to a QTreeView.
    """
    # The source model owns the full device hierarchy and the drag payload for variable leaves.
    model: DynamicsDeviceTreeModel = DynamicsDeviceTreeModel(var_role=var_role, mime_type=mime_type)
    model.setHorizontalHeaderLabels(["Dynamics results"])

    # The invisible root is Qt's insertion point for first-level tree nodes.
    root_item: QtGui.QStandardItem = model.invisibleRootItem()

    device_tpe: DeviceType
    devices_data: Dict[ALL_DEV_TYPES, List[Var]]
    for device_tpe, devices_data in tree_data.items():
        # The first level groups all devices by their type so the tree remains navigable.
        device_type_item: QtGui.QStandardItem = _build_tree_item(text=_get_device_type_label(device_tpe=device_tpe))
        _set_item_icon(item=device_type_item, icon_key=_get_device_type_label(device_tpe=device_tpe))
        root_item.appendRow(device_type_item)

        device: ALL_DEV_TYPES
        variables: List[Var]
        for device, variables in devices_data.items():
            # The second level groups the variables that belong to one physical device.
            device_item: QtGui.QStandardItem = _build_tree_item(text=_get_device_label(device=device))
            device_type_item.appendRow(device_item)

            variable: Var
            for variable in variables:
                # Variable nodes are leaves and store the actual Var object for clicks and drag exports.
                variable_item: QtGui.QStandardItem = _build_tree_item(text=_get_var_label(variable=variable))
                variable_item.setData(variable, var_role)
                device_item.appendRow(variable_item)

    return model


class DynamicsResultsHandler:
    """
    Prepare GUI structures from RMS dynamics results.
    """

    __slots__ = ("results", "tree_data", "tree_model", "proxy_model", "plots_model", "group_idx",
                 "var_role", "group_name_role", "drag_mime_type", "plot_groups")

    def __init__(self, results: RmsResults):
        """
        Build the handler from RMS results data.

        :param results: RMS results container coming from the simulation engine.
        """
        # The original results object is preserved because later GUI actions need arrays and metadata.
        self.results: RmsResults = results

        # These roles are instance-owned so the handler carries all Qt metadata instead of relying on globals.
        self.var_role: int = int(QtCore.Qt.ItemDataRole.UserRole) + 300
        self.group_name_role: int = int(QtCore.Qt.ItemDataRole.UserRole) + 301
        self.drag_mime_type: str = "application/x-veragrid-dynamics-var"

        # Group-name to RMS-group-index mapping is precomputed because plotting needs constant-time access.
        self.group_idx: Dict[str, int] = {str(gr): i for i, gr in enumerate(self.results.rms_events_group_names)}

        # The hierarchical dictionary is the canonical source used to build the device tree.
        self.tree_data: Dict[DeviceType, Dict[ALL_DEV_TYPES, List[Var]]] = self.results.get_devices_dict_tree()

        # The device tree model is the source tree for the left-hand dynamics browser.
        self.tree_model: DynamicsDeviceTreeModel = build_dynamics_tree_model(tree_data=self.tree_data,
                                                                             var_role=self.var_role,
                                                                             mime_type=self.drag_mime_type)

        # The proxy model owns the reversible filtering state used by the device tree view.
        self.proxy_model: QtCore.QSortFilterProxyModel = QtCore.QSortFilterProxyModel()
        self.proxy_model.setSourceModel(self.tree_model)
        self.proxy_model.setRecursiveFilteringEnabled(True)
        self.proxy_model.setFilterCaseSensitivity(QtCore.Qt.CaseSensitivity.CaseInsensitive)
        self.proxy_model.setFilterKeyColumn(0)
        if hasattr(self.proxy_model, "setAutoAcceptChildRows"):
            self.proxy_model.setAutoAcceptChildRows(True)
        else:
            pass

        # Plot groups are stored separately from Qt so CRUD operations are explicit and testable.
        self.plot_groups: DynamicsPlotGroups = DynamicsPlotGroups()

        # The plots model is rebuilt from the domain objects after every CRUD operation.
        self.plots_model: DynamicsPlotsTreeModel = DynamicsPlotsTreeModel(handler=self)
        self.rebuild_plots_model()

    def get_group_name_role(self) -> int:
        """
        Get the Qt role used to store plot-group names in the plots tree.

        :return: Item-data role for plot-group names.
        """
        return self.group_name_role

    def get_drag_mime_type(self) -> str:
        """
        Get the mime type used for dynamics-variable drag-and-drop.

        :return: Mime type string.
        """
        return self.drag_mime_type

    def get_view_model(self) -> QtCore.QSortFilterProxyModel:
        """
        Get the proxy model used by the dynamics device tree view.

        :return: Filter proxy model wrapping the source dynamics tree.
        """
        return self.proxy_model

    def get_plots_model(self) -> DynamicsPlotsTreeModel:
        """
        Get the model used by the dynamics plots tree view.

        :return: Plot-groups tree model.
        """
        return self.plots_model

    def set_search_text(self, search_text: str) -> None:
        """
        Update the proxy-model filter with the given search text.

        :param search_text: Text used to filter the device tree nodes.
        :return: Nothing.
        """
        # Escaping the text keeps the search literal and avoids accidental regular-expression semantics.
        escaped_text: str = QtCore.QRegularExpression.escape(search_text.strip())
        regex: QtCore.QRegularExpression = QtCore.QRegularExpression(escaped_text)
        regex.setPatternOptions(QtCore.QRegularExpression.PatternOption.CaseInsensitiveOption)
        self.proxy_model.setFilterRegularExpression(regex)

    def map_to_source(self, index: QtCore.QModelIndex) -> QtCore.QModelIndex:
        """
        Map a view index from the proxy model back to the source device-tree model.

        :param index: Index received from the tree-view signal.
        :return: Source-model index understood by the handler internals.
        """
        return self.proxy_model.mapToSource(index)

    def get_var_from_index(self, index: QtCore.QModelIndex) -> Var | None:
        """
        Get the RMS variable associated with a clicked device-tree index.

        :param index: Source-model index coming from the dynamics device tree.
        :return: Variable stored in the clicked leaf node, or ``None`` for non-variable nodes.
        """
        if index.isValid():
            item: QtGui.QStandardItem | None = self.tree_model.itemFromIndex(index)
            if item is not None:
                item_data: object = item.data(self.var_role)
                if isinstance(item_data, Var):
                    return item_data
                else:
                    return None
            else:
                return None
        else:
            return None

    def get_plot_group_name_from_index(self, index: QtCore.QModelIndex) -> str | None:
        """
        Get the plot-group name represented by a plots-tree index.

        :param index: Index from the plots tree.
        :return: Group name, or ``None`` when the index does not belong to any group.
        """
        if index.isValid():
            item: QtGui.QStandardItem | None = self.plots_model.itemFromIndex(index)
            if item is not None:
                group_name_data: object = item.data(self.group_name_role)
                if isinstance(group_name_data, str):
                    return group_name_data
                else:
                    parent_item: QtGui.QStandardItem | None = item.parent()
                    if parent_item is not None:
                        parent_group_name_data: object = parent_item.data(self.group_name_role)
                        if isinstance(parent_group_name_data, str):
                            return parent_group_name_data
                        else:
                            return None
                    else:
                        return None
            else:
                return None
        else:
            return None

    def get_group_name_from_drop_index(self, index: QtCore.QModelIndex) -> str | None:
        """
        Resolve the plot-group name targeted by a drop operation.

        :param index: Drop target index.
        :return: Group name, or ``None`` when the drop target is invalid.
        """
        return self.get_plot_group_name_from_index(index=index)

    def get_plot_var_from_index(self, index: QtCore.QModelIndex) -> Var | None:
        """
        Get the variable represented by a plots-tree index.

        :param index: Index from the plots tree.
        :return: Variable when the index points to a variable child, otherwise ``None``.
        """
        if index.isValid():
            item: QtGui.QStandardItem | None = self.plots_model.itemFromIndex(index)
            if item is not None:
                item_data: object = item.data(self.var_role)
                if isinstance(item_data, Var):
                    return item_data
                else:
                    return None
            else:
                return None
        else:
            return None

    def rebuild_plots_model(self) -> None:
        """
        Rebuild the plots-tree model from the plot-group domain objects.

        :return: Nothing.
        """
        # The Qt model is treated as a projection of the handler state so every CRUD operation remains explicit.
        self.plots_model.clear()
        self.plots_model.setHorizontalHeaderLabels(["Dynamic plots"])

        root_item: QtGui.QStandardItem = self.plots_model.invisibleRootItem()
        group: DynamicsPlotGroup
        for group in self.plot_groups.get_groups():
            group_item: QtGui.QStandardItem = _build_tree_item(text=group.get_name())
            group_item.setData(group.get_name(), self.group_name_role)
            _set_item_icon(item=group_item, icon_key="Dynamic")
            root_item.appendRow(group_item)

            variable: Var
            for variable in group.get_vars():
                variable_item: QtGui.QStandardItem = _build_tree_item(text=_get_var_label(variable=variable))
                variable_item.setData(variable, self.var_role)
                group_item.appendRow(variable_item)

    def _build_next_group_name(self) -> str:
        """
        Build the next default plot-group name.

        :return: New plot-group name that does not clash with existing ones.
        """
        group_number: int = 1
        while self.plot_groups.get_group(name="Plot " + str(group_number)) is not None:
            group_number += 1
        return "Plot " + str(group_number)

    def get_next_group_name(self) -> str:
        """
        Get the next available default plot-group name.

        :return: Suggested group name.
        """
        return self._build_next_group_name()

    def create_plot_group(self, name: str) -> bool:
        """
        Create a plot group and refresh the plots tree.

        :param name: Requested plot-group name.
        :return: ``True`` when the group was created.
        """
        created: bool = self.plot_groups.create_group(name=name)
        if created:
            self.rebuild_plots_model()
            return True
        else:
            return False

    def delete_plot_group(self, group_name: str) -> bool:
        """
        Delete one plot group and refresh the plots tree.

        :param group_name: Plot-group name.
        :return: ``True`` when the group existed and was deleted.
        """
        deleted: bool = self.plot_groups.delete_group(name=group_name)
        if deleted:
            self.rebuild_plots_model()
            return True
        else:
            return False

    def add_var_to_group(self, group_name: str, var_uid: int) -> bool:
        """
        Add one variable to a plot group and refresh the plots tree.

        :param group_name: Target plot-group name.
        :param var_uid: Variable uid obtained from the drag payload.
        :return: ``True`` when the variable was inserted.
        """
        group: DynamicsPlotGroup | None = self.plot_groups.get_group(name=group_name)
        variable: Var | None = self.results.get_var(uid=var_uid)
        if group is not None:
            if variable is not None:
                inserted: bool = group.add_var(variable=variable)
                if inserted:
                    self.rebuild_plots_model()
                    return True
                else:
                    return False
            else:
                return False
        else:
            return False

    def remove_var_from_group(self, group_name: str, var_uid: int) -> bool:
        """
        Remove one variable from a plot group and refresh the plots tree.

        :param group_name: Plot-group name.
        :param var_uid: Variable uid.
        :return: ``True`` when the variable was removed.
        """
        group: DynamicsPlotGroup | None = self.plot_groups.get_group(name=group_name)
        variable: Var | None = self.results.get_var(uid=var_uid)
        if group is not None:
            if variable is not None:
                removed: bool = group.remove_var(variable=variable)
                if removed:
                    self.rebuild_plots_model()
                    return True
                else:
                    return False
            else:
                return False
        else:
            return False

    def delete_plot_entry_from_index(self, index: QtCore.QModelIndex) -> bool:
        """
        Delete the selected group or variable from the plots tree.

        :param index: Selected plots-tree index.
        :return: ``True`` when something was deleted.
        """
        selected_var: Var | None = self.get_plot_var_from_index(index=index)
        if selected_var is not None:
            group_name: str | None = self.get_plot_group_name_from_index(index=index)
            if group_name is not None:
                return self.remove_var_from_group(group_name=group_name, var_uid=selected_var.uid)
            else:
                return False
        else:
            group_name = self.get_plot_group_name_from_index(index=index)
            if group_name is not None:
                return self.delete_plot_group(group_name=group_name)
            else:
                return False

    def plot_var(self, var: Var, group_name: str) -> None:
        """
        Plot one variable for one RMS events group.

        :param var: Variable to plot.
        :param group_name: RMS events group name.
        :return: Nothing.
        """
        gr_idx: int = self.group_idx[group_name]
        self.results.plot_var(var=var, group_idx=gr_idx)

    def plot_group(self, plot_group_name: str, rms_group_name: str) -> bool:
        """
        Plot all variables stored in one plot group.

        :param plot_group_name: Plot-group name selected by the user.
        :param rms_group_name: RMS events group name selected in the combobox.
        :return: ``True`` when the plot group existed and was plotted.
        """
        plot_group: DynamicsPlotGroup | None = self.plot_groups.get_group(name=plot_group_name)
        if plot_group is not None:
            variables: List[Var] = plot_group.get_vars()
            if len(variables) > 0:
                group_idx: int = self.group_idx[rms_group_name]
                figure = plt.figure(figsize=(12, 8))
                axis = figure.add_subplot(111)

                variable: Var
                for variable in variables:
                    values_idx: int = self.results.uid2idx[variable.uid]
                    y_values = self.results.values[:, values_idx, group_idx]
                    axis.plot(self.results.time_array, y_values, label=variable.name)

                axis.legend()
                axis.set_title(plot_group_name + " - " + rms_group_name)
                plt.show()
                return True
            else:
                return False
        else:
            return False

    def plot_entry_from_index(self, index: QtCore.QModelIndex, rms_group_name: str) -> bool:
        """
        Plot the selected plots-tree entry.

        :param index: Selected plots-tree index.
        :param rms_group_name: Selected RMS events group name.
        :return: ``True`` when something was plotted.
        """
        selected_var: Var | None = self.get_plot_var_from_index(index=index)
        if selected_var is not None:
            self.plot_var(var=selected_var, group_name=rms_group_name)
            return True
        else:
            plot_group_name: str | None = self.get_plot_group_name_from_index(index=index)
            if plot_group_name is not None:
                return self.plot_group(plot_group_name=plot_group_name, rms_group_name=rms_group_name)
            else:
                return False

    def get_data_from_plot_index(self, index: QtCore.QModelIndex, rms_group_name: str) -> ResultsModel | None:
        """

        :param index:
        :param rms_group_name:
        :return:
        """
        plot_group_name: str | None = self.get_plot_group_name_from_index(index=index)

        if plot_group_name is not None:

            plot_group: DynamicsPlotGroup | None = self.plot_groups.get_group(name=plot_group_name)
            variables: List[Var] = plot_group.get_vars()
            if len(variables) > 0:
                group_idx: int = self.group_idx[rms_group_name]
                data = self.results.get_vars_data(var_list=variables, group_idx=group_idx)

                var_names = np.array([v.name for v in variables])

                table = ResultsTable(
                    data=data,
                    index=self.results.time_array,
                    columns=var_names,
                    title=plot_group_name,
                    ylabel="",
                    cols_device_type=DeviceType.NoDevice,
                    idx_device_type=DeviceType.NoDevice
                )

                mdl = ResultsModel(table=table)

                return mdl
            else:
                return None
        else:
            return None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import io
import sys
from typing import Any
import code
import contextlib
import rlcompleter
import traceback

from PySide6.QtCore import Qt, Signal, QObject
from PySide6.QtGui import QTextCursor, QFont
from PySide6.QtWidgets import QPlainTextEdit
from PySide6.QtWidgets import QApplication, QMainWindow
from PySide6.QtWidgets import QCompleter
from PySide6.QtCore import QStringListModel
from VeraGrid.Gui.python_highlighter import PythonHighlighter
from VeraGrid.Gui.font_config import CONSOLE_TEXT_SIZE


class _GuiOutput(QObject):
    written = Signal(str)


class ConsoleInterpreter(code.InteractiveInterpreter):
    def __init__(self, locals=None, write_callback=None):
        super().__init__(locals=locals or {})
        self._write_callback = write_callback

    def write(self, data: str):
        # InteractiveInterpreter uses this for tracebacks, etc.
        if self._write_callback:
            self._write_callback(data)

    def showtraceback(self):
        tb = traceback.format_exc()
        self.write(tb)


class _BufferStream(io.TextIOBase):
    """File-like stream to capture stdout/stderr during execution."""

    def __init__(self, write_fn):
        super().__init__()
        self._write_fn = write_fn

    def write(self, s):
        self._write_fn(s)
        return len(s)

    def flush(self):
        pass


class PythonConsole(QPlainTextEdit):
    PROMPT_PRIMARY = ">>> "
    PROMPT_SECONDARY = "... "

    def __init__(self, banner="", locals=None, parent=None):
        """

        :param banner:
        :param locals:
        :param parent:
        """
        super().__init__(parent)

        font = QFont("Consolas", CONSOLE_TEXT_SIZE)
        self.setFont(font)
        self.setUndoRedoEnabled(False)
        self.setLineWrapMode(QPlainTextEdit.LineWrapMode.NoWrap)

        PythonHighlighter(self.document())

        self._history = []
        self._history_index = 0
        self._buffer = []  # multiline buffer
        self._input_start_pos = 0

        # GUI-thread output marshalling
        self._gui_out = _GuiOutput()
        self._gui_out.written.connect(self._append_output)

        self._interpreter = ConsoleInterpreter(
            locals=locals,
            write_callback=self._emit_output,
        )
        self._completer = QCompleter(self)
        self._completer.setWidget(self)
        self._completer.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)
        self._completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self._completer.activated.connect(self._insert_completion)

        if banner:
            self._append_output(banner + "\n")

        self._insert_prompt(primary=True)
        self.cursorPositionChanged.connect(self._enforce_cursor)

    def reset(self):
        self.document().clear()
        self._insert_prompt(primary=True)
        self.cursorPositionChanged.connect(self._enforce_cursor)

    def add_var(self, name: str, val: Any) -> None:
        """
        Add variable to the interpreter
        :param name: name of the variable
        :param val: value or pointer
        """
        self._interpreter.locals[name] = val

    def append_output(self, text: str):
        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.MoveOperation.End)
        cursor.insertText(text)
        self.setTextCursor(cursor)

        # After output, input must start at the end of document
        self._input_start_pos = self.document().characterCount() - 1

    def execute(self, command: str):
        """
        Run a command and display output.
        """

        try:
            old_stdout = sys.stdout
            old_stderr = sys.stderr
            sys.stdout = io.StringIO()
            sys.stderr = io.StringIO()

            success = self._interpreter.runcode(command)
            stdout_output = sys.stdout.getvalue()
            stderr_output = sys.stderr.getvalue()
            sys.stdout = old_stdout
            sys.stderr = old_stderr

            if stdout_output:
                self.append_output(stdout_output)
            if stderr_output:
                self.append_output(stderr_output)

        except Exception as e:
            self.append_output(str(e))

    # ----------------------------
    # Prompt + input boundary
    # ----------------------------
    def _safe_input_start_pos(self) -> int:
        doc = self.document()
        # characterCount() includes a trailing '\n'
        return min(self._input_start_pos, doc.characterCount() - 1)

    def _insert_prompt(self, primary: bool):
        prompt = self.PROMPT_PRIMARY if primary else self.PROMPT_SECONDARY
        self.moveCursor(QTextCursor.MoveOperation.End)
        self.insertPlainText(prompt)
        self._input_start_pos = self.textCursor().position()
        self.moveCursor(QTextCursor.MoveOperation.End)

    def _current_input(self) -> str:
        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.MoveOperation.StartOfLine, QTextCursor.MoveMode.KeepAnchor)
        line = cursor.selectedText()
        # Remove prompt prefix
        if line.startswith(self.PROMPT_PRIMARY):
            return line[len(self.PROMPT_PRIMARY):]
        if line.startswith(self.PROMPT_SECONDARY):
            return line[len(self.PROMPT_SECONDARY):]
        return line

    def _replace_current_line(self, text: str):
        cursor = self.textCursor()

        cursor.movePosition(QTextCursor.MoveOperation.StartOfLine)
        cursor.movePosition(QTextCursor.MoveOperation.EndOfLine, QTextCursor.MoveMode.KeepAnchor)
        cursor.removeSelectedText()

        prompt = self.PROMPT_SECONDARY if self._buffer else self.PROMPT_PRIMARY
        cursor.insertText(prompt + text)

        # Recompute input boundary *after prompt only*
        cursor.movePosition(QTextCursor.MoveOperation.StartOfLine)
        cursor.movePosition(QTextCursor.MoveOperation.Right, QTextCursor.MoveMode.MoveAnchor, len(prompt))
        self._input_start_pos = cursor.position()

        self.setTextCursor(cursor)

    # ----------------------------
    # Thread-safe output
    # ----------------------------

    def _emit_output(self, text: str):
        # Always go through a signal (safe even if called from worker threads)
        self._gui_out.written.emit(text)

    def _append_output(self, text: str):
        self.moveCursor(QTextCursor.MoveOperation.End)
        self.insertPlainText(text)
        self.moveCursor(QTextCursor.MoveOperation.End)

    # ----------------------------
    # Hard prompt protection
    # ----------------------------

    def _enforce_cursor(self):
        safe_pos = self._safe_input_start_pos()
        c = self.textCursor()
        if c.position() < safe_pos:
            c.setPosition(safe_pos)
            self.setTextCursor(c)

    def _selection_crosses_prompt(self) -> bool:
        c = self.textCursor()
        return c.hasSelection() and c.selectionStart() < self._input_start_pos

    def _accept_completion_from_popup(self):
        popup = self._completer.popup()
        index = popup.currentIndex()
        if not index.isValid():
            popup.hide()
            return

        completion = index.data()

        # Save cursor position *before* modifying text
        cursor = self.textCursor()
        cursor.beginEditBlock()

        self._replace_current_line(completion)

        cursor.endEditBlock()

        # IMPORTANT: restore cursor explicitly to end of inserted text
        self.moveCursor(QTextCursor.MoveOperation.End)

        popup.hide()

    # ----------------------------
    # Key handling
    # ----------------------------

    def keyPressEvent(self, event):
        # -------------------------------------------------
        # If autocomplete popup is visible → accept / close
        # -------------------------------------------------
        popup = self._completer.popup()
        if popup.isVisible():
            if event.key() in (Qt.Key.Key_Tab, Qt.Key.Key_Return, Qt.Key.Key_Enter):
                # Accept the current completion
                self._accept_completion_from_popup()
                return

            if event.key() == Qt.Key.Key_Escape:
                # Close the popup
                popup.hide()
                return

            # Let the popup handle navigation keys (Up/Down/PageUp/PageDown)
            if event.key() in (Qt.Key.Key_Up, Qt.Key.Key_Down, Qt.Key.Key_PageUp, Qt.Key.Key_PageDown):
                return super().keyPressEvent(event)

            # Prevent Left and Right Arrow keys from inserting characters
            if event.key() in (Qt.Key.Key_Left, Qt.Key.Key_Right):
                return  # Don't allow Left/Right Arrow to insert text while popup is visible

            # Update the completion model as text changes (only when popup is visible)
            self._autocomplete()

        # -------------------------------------------------
        # Insert tab space or trigger completion when popup is NOT visible
        # -------------------------------------------------
        elif event.key() == Qt.Key.Key_Tab:
            # If no popup is visible, insert a tab character (space or actual tab)
            cursor = self.textCursor()
            cursor.insertText("\t")  # Insert tab space
            self.setTextCursor(cursor)
            return

        # Trigger completion popup when Ctrl + Space is pressed
        elif event.key() == Qt.Key.Key_Space and event.modifiers() & Qt.KeyboardModifier.ControlModifier:
            self._autocomplete()
            return

        c = self.textCursor()

        # Prevent any edits that would touch the prompt or earlier output
        if self._selection_crosses_prompt():
            if event.key() in (Qt.Key.Key_Backspace, Qt.Key.Key_Delete) or event.text():
                return

        if event.key() == Qt.Key.Key_Backspace:
            safe_pos = self._safe_input_start_pos()
            if c.position() <= safe_pos:
                return

        if event.key() == Qt.Key.Key_Delete:
            safe_pos = self._safe_input_start_pos()
            if c.position() < safe_pos:
                return

        if event.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            self._handle_enter()
            return

        if event.key() == Qt.Key.Key_Up:
            self._history_prev()
            return

        if event.key() == Qt.Key.Key_Down:
            self._history_next()
            return

        if event.key() == Qt.Key.Key_Tab:
            self._autocomplete()
            return

        super().keyPressEvent(event)

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        self._enforce_cursor()

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        self._enforce_cursor()

    def insertFromMimeData(self, source):
        # Block paste that would overwrite prompt/output
        if self._selection_crosses_prompt():
            return
        super().insertFromMimeData(source)

    # ----------------------------
    # Execution (safe redirection)
    # ----------------------------

    def _handle_enter(self):
        line = self._current_input()
        self._history.append(line)
        self._history_index = len(self._history)

        self._buffer.append(line)
        source = "\n".join(self._buffer)

        self._append_output("\n")

        out_stream = _BufferStream(self._emit_output)
        err_stream = _BufferStream(self._emit_output)

        # Redirect ONLY during execution; do NOT touch global sys.stdout permanently.
        with contextlib.redirect_stdout(out_stream), contextlib.redirect_stderr(err_stream):
            more = self._interpreter.runsource(source)

        if more:
            self._insert_prompt(primary=False)
        else:
            self._buffer.clear()
            self._insert_prompt(primary=True)

    # ----------------------------
    # History
    # ----------------------------

    def _history_prev(self):
        if not self._history:
            return
        self._history_index = max(0, self._history_index - 1)
        self._replace_current_line(self._history[self._history_index])

    def _history_next(self):
        if not self._history:
            return
        self._history_index = min(len(self._history), self._history_index + 1)
        text = "" if self._history_index == len(self._history) else self._history[self._history_index]
        self._replace_current_line(text)

    # ----------------------------
    # Autocomplete
    # ----------------------------

    def _insert_completion(self, completion: str):
        self._replace_current_line(completion)

    def _autocomplete(self):
        text = self._current_input()
        if not text:
            return

        completer = rlcompleter.Completer(self._interpreter.locals)

        matches = list()
        i = 0
        while True:
            m = completer.complete(text, i)
            if m is None:
                break
            matches.append(m)
            i += 1

        if not matches:
            return

        if len(matches) == 1:
            self._replace_current_line(matches[0])
            return

        # Multiple matches → popup
        model = QStringListModel(matches, self._completer)
        self._completer.setModel(model)

        cursor = self.textCursor()
        cr = self.cursorRect(cursor)
        cr.setWidth(self._completer.popup().sizeHintForColumn(0) +
                    self._completer.popup().verticalScrollBar().sizeHint().width())

        self._completer.complete(cr)


if __name__ == "__main__":
    import numpy as np

    class ConsoleMainWindow(QMainWindow):
        def __init__(self):
            super().__init__()
            self.setWindowTitle("PySide6 Python Console")
            console = PythonConsole(banner="Welcome to Python Console!")
            self.setCentralWidget(console)

            console.add_var("np", np)


    app = QApplication(sys.argv)
    window = ConsoleMainWindow()
    window.resize(800, 600)
    window.show()
    sys.exit(app.exec())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from PySide6 import QtCore, QtWidgets


class WrappableTableModel(QtCore.QAbstractTableModel):
    """
    Class to populate a Qt table view with the properties of objects
    """

    def __init__(self, parent: QtWidgets.QTableView = None):
        """

        :param parent: Parent object: the QTableView object
        """
        QtCore.QAbstractTableModel.__init__(self, parent)

        # flag for the headers text wrapper: HeaderViewWithWordWrap
        self._hide_headers_mode = False

    def hide_headers(self) -> None:
        """

        :return:
        """
        self._hide_headers_mode = True

    def unhide_headers(self) -> None:
        """

        :return:
        """
        self._hide_headers_mode = False



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations

from PySide6 import QtWidgets
from PySide6.QtCore import Qt

from VeraGrid.Gui.GridMerge.build_diff_tree import populate_tree
from VeraGrid.Gui.GridMerge.grid_merge_gui import Ui_Dialog
from VeraGrid.Gui.general_dialogues import LogsDialogue
from VeraGridEngine.basic_structures import Logger
from VeraGridEngine.Devices.multi_circuit import MultiCircuit


class GridMergeDialogue(QtWidgets.QDialog):
    """
    GridMergeDialogue
    """

    def __init__(self, grid: MultiCircuit, diff: MultiCircuit):
        """

        :param grid:
        """
        QtWidgets.QDialog.__init__(self)
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Grid merges & acquisitions')

        self.ui.treeWidget.setHeaderLabels(
            ["Grid", "Object type", "action", "idtag", "name", "property", "value", "new value"]
        )

        self.logger = Logger()

        self._base_grid: MultiCircuit = grid

        self.all_elms_base_dict, ok = self._base_grid.get_all_elements_dict(logger=self.logger)

        if not ok:
            dlg = LogsDialogue('The base circuit has duplicated idtags and cannot be merged :(', self.logger)
            dlg.exec()
            return

        self._diff: MultiCircuit = diff

        self.diff_objects_dict, ok2 = self._diff.get_all_elements_dict(logger=self.logger)

        if not ok2:
            dlg = LogsDialogue('The diff circuit has duplicated idtags and cannot be merged :(', self.logger)
            dlg.exec()
            return

        self.build_tree()

        self.added_grid: bool = False
        self.merged_grid: bool = False

        # tree item changes
        self.ui.treeWidget.itemChanged.connect(self.handle_item_changed)

        self.ui.acceptButton.clicked.connect(self.merge_grid)

        self.ui.addButton.clicked.connect(self.add_grid)

    def handle_item_changed(self, item: QtWidgets.QTreeWidgetItem, column: int):
        """

        :param item:
        :param column: THis must be here because this is an event handler
        :return:
        """
        if item.parent() is None:  # Root item

            if item.parent() is None:
                # the item is a parent
                idtag = item.text(3)
                elm = self.diff_objects_dict.get(idtag)

                # set the "selected to merge" thing from the first column
                elm.selected_to_merge = item.checkState(0) == Qt.CheckState.Checked

                # set the check status of the children to the same as the parent
                state = item.checkState(0)
                for i in range(item.childCount()):
                    item.child(i).setCheckState(0, state)

                print(f"obj: {elm.name}.selected_to_merge={elm.selected_to_merge}")

            else:
                # if it is a child, then it represents the property of an object
                parent_item = item.parent()
                idtag = parent_item.text(3)
                elm = self.diff_objects_dict.get(idtag)
                prop_name = item.text(4)
                prop = elm.get_property_by_name(prop_name=prop_name)

                # we want to mark each property as selectable or not
                prop.selected_to_merge = item.checkState(0) == Qt.CheckState.Checked

                print(f"prop: {elm.name}.{prop_name}.selected_to_merge={prop.selected_to_merge}")

    def set_diff(self, diff: MultiCircuit):
        """

        :param diff:
        :return:
        """
        self._diff = diff

    def build_tree(self):
        """
        Build the display tree
        """

        all_no_action = populate_tree(tree_widget=self.ui.treeWidget,
                                      base=self._base_grid,
                                      diff=self._diff,
                                      all_elms_base_dict=self.all_elms_base_dict)

        # if all_no_action: we cannot merge
        self.ui.acceptButton.setVisible(not all_no_action)

    def merge_grid(self):
        """
        Apply the changes
        :return:
        """

        self._base_grid.merge_circuit(new_grid=self._diff)

        # Done
        self.added_grid: bool = False
        self.merged_grid: bool = True
        self.close()

    def add_grid(self):
        """
        The elements of the grid will be added with new idtags.
        This is useful in the case you want to compose a new grid from grids that are the same.
        :return:
        """
        self._base_grid.add_circuit(self._base_grid)
        self.added_grid: bool = True
        self.merged_grid: bool = False
        self.close()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations
import os
import sys
import importlib
import importlib.util
import hashlib
from typing import List, Dict, TYPE_CHECKING, Callable
import json
import zipfile
import shutil
from PySide6.QtGui import QPixmap

from VeraGridEngine.IO.file_system import plugins_path
from VeraGrid.__version__ import __VeraGrid_VERSION__

if TYPE_CHECKING:
    from VeraGrid.Gui.Main.SubClasses.Settings.configuration import ConfigurationMain


class PluginFunction:
    """
    Class to handle external function pointers
    """

    def __init__(self) -> None:

        self.name = ""
        self.alias = ""
        self.call_gui = False
        self.function_ptr = None

    def get_pointer_lambda(self, gui_instance: ConfigurationMain) -> Callable:
        """
        Really hard core magic to avoid lambdas shadow each other due to late binding

        lambda e=True, func=self.function_ptr: func(self)

        explanation:
        - e is a bool parameter that the QAction sends when triggered,
            set it by default True, otherwise fails on windows
        - func=self.function_ptr is there for the lambda to force the usage of the value of self.function_ptr
            during the iteration and not after the loop since lambdas in a loop are lazy evaluated
        - func(self) is then what I wanted to lambda in the first place
        """
        return lambda e=True, func=self.function_ptr: func(gui_instance)

    def to_dict(self) -> Dict[str, str]:
        """
        To dict
        :return: string, string dictionary to save in json format
        """
        return {
            "name": self.name,
            "alias": self.alias,
            "call_gui": self.call_gui,
        }

    def parse(self, data: Dict[str, str]) -> None:
        """
        Parse data
        :param data: Data like the one saved
        """
        self.name = data.get('name', '').strip()
        self.alias = data.get('alias', '').strip()
        self.call_gui = data.get('call_gui', False)

    def read_plugin(self, plugin_path: str) -> None:
        """
        Read the pointed plugin
        :param plugin_path: plugin file path
        """

        # hot read python file˘
        if self.name != "":
            try:

                # read the main function (the one launched upon click on the plugin)
                self.function_ptr = load_function_from_file_path(
                    file_path=plugin_path,
                    function_name=self.name
                )

            except ImportError as e:
                print(e)
            except AttributeError as e:
                print(e)
            except TypeError as e:
                print(e)


class PluginInfo:
    """
    Plugin information
    """

    def __init__(self, folder, file_path) -> None:
        """

        """
        # file info
        self.folder = folder
        self.file_path = file_path

        # plugin data
        self.name = ""
        self.code_file_path = ""
        self.icon_path = ""
        self.version = "0.0.0"
        self.veragrid_version = "5.4.0"

        self.main_fcn: PluginFunction = PluginFunction()

        self.investments_fcn: PluginFunction = PluginFunction()

        self.icon: QPixmap | None = None

        if os.path.exists(file_path):
            with open(file_path, 'r') as f:
                data = json.load(f)
                self.parse(data)
                self.read_plugin()

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name

    def to_dict(self) -> Dict[str, str | Dict[str, str]]:
        """
        To dict
        :return: string, string dictionary to save in json format
        """
        return {
            "name": self.name,
            "version": self.version,
            "veragrid_version": self.veragrid_version,
            "path": self.code_file_path,
            "icon_path": self.icon_path,
            "main_fcn": self.main_fcn.to_dict(),
            "investments_fcn": self.investments_fcn.to_dict(),
        }

    def parse(self, data: Dict[str, str | Dict[str, str]]) -> None:
        """
        Parse data
        :param data: Data like the one saved
        """
        self.name = data.get('name', '')
        self.version = data.get('version', '0.0.0')
        self.veragrid_version = data.get('veragrid_version', '5.2.0')
        self.code_file_path = data.get('path', '')
        self.icon_path = data.get('icon_path', '')

        if 'main_fcn' in data.keys():
            self.main_fcn.parse(data['main_fcn'])

        if 'investments_fcn' in data.keys():
            self.investments_fcn.parse(data['investments_fcn'])

    def read_plugin(self) -> None:
        """
        Read the pointed plugin
        """
        plugin_path = os.path.join(self.folder, self.code_file_path)

        if os.path.exists(plugin_path):

            if plugin_path.endswith('.py'):

                # hot read python file searching for the functions declared
                self.main_fcn.read_plugin(plugin_path)
                self.investments_fcn.read_plugin(plugin_path)

            else:
                print(f"Plugin {self.name}: Path {plugin_path} not a python file :(")
        else:
            print(f"Plugin {self.name}: Path {plugin_path} not found :/")

        # Read the icon file if available
        icon_path = os.path.join(self.folder, self.icon_path)
        if os.path.exists(icon_path):
            self.icon = QPixmap(icon_path)
        else:
            print(f"Plugin {self.name}: Path {icon_path} not found :/")

    def is_greater(self, other: "PluginInfo") -> int:
        """

        :param other:
        :return:
        """
        v1_tuple = tuple(map(int, self.version.split(".")))
        v2_tuple = tuple(map(int, other.version.split(".")))

        # Compare the tuples
        if v1_tuple < v2_tuple:
            return 1
        elif v1_tuple > v2_tuple:
            return -1
        else:
            return 0

    def is_compatible(self) -> int:
        """
        Check if the plugin is compatible
        :return:
        """
        v1_tuple = tuple(map(int, __VeraGrid_VERSION__.split(".")))
        v2_tuple = tuple(map(int, self.veragrid_version.split(".")))

        # Compare the tuples
        if v1_tuple < v2_tuple:
            return False
        else:
            return True


class PluginsInfo:
    """
    Plugins information
    """

    def __init__(self) -> None:
        """

        """
        # add the plugins directory to the pythonpath
        sys.path.insert(0, plugins_path())

        self.plugins: Dict[str, PluginInfo] = dict()
        self.read()

    def to_data(self) -> List[Dict[str, str]]:
        """
        Get dictionary of plugin data
        :return:
        """
        return [pl.to_dict() for key, pl in self.plugins.items()]

    def read(self) -> None:
        """
        Thead the plugins info file
        :return:
        """
        self.plugins.clear()

        # List all JSON files in the folder
        folder = plugins_path()
        # for file in os.listdir(folder):
        for subdir, dirs, files in os.walk(folder):
            for file in files:
                if file.endswith('.plugin.json'):
                    file_path = os.path.join(subdir, file)

                    try:
                        pl = PluginInfo(subdir, file_path)
                        self.plugins[pl.name] = pl

                    except json.JSONDecodeError as e:
                        print("Error reading the plugins index", e)


def load_function_from_file_path(file_path: str, function_name: str):
    """
    Dynamically load a function from a Python file at a given file path.

    :param file_path: The path to the Python (.py) file.
    :param function_name: The name of the function to load from the file.
    :return: The loaded function object.
    :raises FileNotFoundError: If the specified file does not exist.
    :raises ImportError: If the module cannot be imported.
    :raises AttributeError: If the function does not exist in the module.
    :raises TypeError: If the retrieved attribute is not callable.
    """
    # Ensure the file exists
    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"No such file: {file_path}")

    # Generate a unique module name to avoid conflicts
    # Here, we use the file's absolute path hashed to ensure uniqueness
    absolute_path = os.path.abspath(file_path)
    module_name = f"dynamic_module_{hashlib.md5(absolute_path.encode()).hexdigest()}"

    # Create a module specification from the file location
    spec = importlib.util.spec_from_file_location(module_name, absolute_path)
    if spec is None:
        raise ImportError(f"Cannot create a module spec for '{file_path}'")

    # Create a new module based on the spec
    module = importlib.util.module_from_spec(spec)

    spec.loader.exec_module(module)

    try:
        # Execute the module to populate its namespace
        spec.loader.exec_module(module)
    except Exception as e:
        raise ImportError(f"Failed to execute module '{file_path}': {e}") from e

    # Retrieve the function from the module
    if not hasattr(module, function_name):
        raise AttributeError(f"The function '{function_name}' does not exist in '{file_path}'")

    func = getattr(module, function_name)

    if not callable(func):
        raise TypeError(f"'{function_name}' in '{file_path}' is not callable")

    return func


def pack_plugin(name: str,
                pkg_folder: str,
                python_file: str,
                main_name: str,
                icon_file: str,
                version: str,
                call_gui: bool,
                veragrid_version: str = "5.2.10"):
    """
    Create plugin package
    :param name: Name of the plugin
    :param pkg_folder: Source folder of the plugin
    :param python_file: main python file for the plugin (relative to pkg_folder)
    :param main_name: name of the main function within the python_file
    :param icon_file: icon file (relative to pkg_folder)
    :param version: Version of the plugin
    :param call_gui: does the main function
    :param veragrid_version: veragrid version of the plugin
    :return: final name of the plugin
    """
    plugin_data = {
        "plugins_tech_version": "1.0.0",
        "name": name,
        "path": python_file,
        "icon_path": icon_file,
        "version": version,
        "veragrid_version": veragrid_version,
        "main_fcn": {
            "name": main_name,
            "alias": name,
            "call_gui": call_gui
        }
    }

    v2 = version.replace(".", "_")
    filename_zip = f'{name}_{v2}.vgplugin'
    with zipfile.ZipFile(filename_zip, 'w', zipfile.ZIP_DEFLATED) as f_zip_ptr:

        folder_name = os.path.basename(pkg_folder)

        config_file = os.path.join(folder_name, "config.plugin.json")

        f_zip_ptr.writestr("manifest.json", json.dumps({
            "name": name,
            "folder": folder_name,
            "version": version,
            "config_file": config_file
        }))

        # save the config files
        f_zip_ptr.writestr(config_file, json.dumps(plugin_data))

        # Get the parent directory of the folder
        parent_folder = os.path.dirname(pkg_folder)

        for root, dirs, files in os.walk(pkg_folder):
            for file in files:
                file_path = os.path.join(root, file)
                # Add file to the zip archive, preserving folder structure
                f_zip_ptr.write(file_path, os.path.relpath(file_path, parent_folder))

    return filename_zip


def get_plugin_info(plugin_file: str) -> PluginInfo | None:
    with zipfile.ZipFile(plugin_file, 'r') as zipf:

        if "manifest.json" in zipf.namelist():
            # read the manifest
            with zipf.open("manifest.json") as json_file:
                data = json.load(json_file)
                info = PluginInfo("", "")
                info.parse(data)
                return info
        else:
            return None


def install_plugin(plugin_file: str):
    """

    :param plugin_file:
    :return:
    """

    plugins_pth = plugins_path()

    with zipfile.ZipFile(plugin_file, 'r') as zipf:

        # read the manifest
        with zipf.open("manifest.json") as json_file:
            data = json.load(json_file)

            folder_name = data.get("folder", None)

            dst_folder = os.path.join(plugins_pth, folder_name)
            if os.path.exists(dst_folder):
                shutil.rmtree(dst_folder)

        if folder_name is not None:
            for member in zipf.namelist():
                if member.startswith(f"{folder_name}/"):  # Replace with the folder you want to extract
                    zipf.extract(member, plugins_pth)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import sys
from PySide6 import QtWidgets
from typing import List

from VeraGrid.Gui.ContingencyPlanner.contingency_planner_gui import Ui_MainWindow
import VeraGridEngine.Devices as dev
from VeraGridEngine.enumerations import DeviceType
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
import VeraGrid.Gui.gui_functions as gf
from VeraGridEngine.Simulations.ContingencyAnalysis.contingency_plan import generate_automatic_contingency_plan


class ContingencyPlannerGUI(QtWidgets.QDialog):
    """
    ContingencyPlannerGUI
    """

    def __init__(self, parent=None, grid: MultiCircuit = None):
        """

        :param parent:
        """
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.setWindowTitle('Contingency planner')

        self.circuit = grid

        self.contingency_branch_types = [DeviceType.LineDevice,
                                         DeviceType.DCLineDevice,
                                         DeviceType.Transformer2WDevice,
                                         DeviceType.VscDevice,
                                         DeviceType.UpfcDevice]

        self.contingency_injection_types = [DeviceType.GeneratorDevice,
                                            DeviceType.BatteryDevice]

        self.ui.contingencyBranchTypesListView.setModel(gf.get_list_model(self.contingency_branch_types,
                                                                          checks=True, check_value=True))

        self.ui.contingenctyInjectionsListView.setModel(gf.get_list_model(self.contingency_injection_types,
                                                                          checks=True, check_value=True))

        # contingencies
        self.contingencies: List[dev.Contingency] = list()
        self.contingency_groups: List[dev.ContingencyGroup] = list()

        self.ui.autoNminusXButton.clicked.connect(self.auto_generate_contingencies)

        self.generated_results = False

    def msg(self, text, title="Warning"):
        """
        Message box
        :param text: Text to display
        :param title: Name of the window
        """
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Icon.Information)
        msg.setText(text)
        # msg.setInformativeText("This is additional information")
        msg.setWindowTitle(title)
        # msg.setDetailedText("The details are as follows:")
        msg.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
        retval = msg.exec()

    def auto_generate_contingencies(self):
        """
        Automatically generate the contingency plan from the selection
        :return:
        """

        # filters
        branch_indices = gf.get_checked_indices(self.ui.contingencyBranchTypesListView.model())
        injection_indices = gf.get_checked_indices(self.ui.contingenctyInjectionsListView.model())

        branch_types = [self.contingency_branch_types[i] for i in branch_indices]
        injection_types = [self.contingency_injection_types[i] for i in injection_indices]

        # generate the contingency plan
        self.contingencies, self.contingency_groups = generate_automatic_contingency_plan(
            grid=self.circuit,
            k=self.ui.contingencyNspinBox.value(),
            consider_branches=self.ui.addBranchesToContingencyCheckBox.isChecked(),
            filter_branches_by_voltage=self.ui.filterContingencyBranchesByVoltageCheckBox.isChecked(),
            vmin=self.ui.filterContingencyBranchesByVoltageMinSpinBox.value(),
            vmax=self.ui.filterContingencyBranchesByVoltageMaxSpinBox.value(),
            branch_types=branch_types,
            consider_injections=self.ui.addInjectionsToContingencyCheckBox.isChecked(),
            filter_injections_by_power=self.ui.contingencyFilterInjectionsByPowerCheckBox.isChecked(),
            contingency_perc=self.ui.contingencyInjectionPowerReductionSpinBox.value(),
            pmin=self.ui.contingencyFilterInjectionsByPowerMinSpinBox.value(),
            pmax=self.ui.contingencyFilterInjectionsByPowerMaxSpinBox.value(),
            injection_types=injection_types
        )

        self.generated_results = True
        self.close()


# if __name__ == "__main__":
#     app = QtWidgets.QApplication(sys.argv)
#     window = ContingencyPlannerGUI()
#     window.resize(1.61 * 700.0, 600.0)  # golden ratio
#     window.show()
#     sys.exit(app.exec())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from typing import List
from PySide6.QtCore import QThread, Signal
from VeraGridEngine.IO.veragrid.results_export import export_drivers
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.Simulations.types import DRIVER_OBJECTS
from VeraGridEngine.basic_structures import Logger


class ExportAllThread(QThread):
    """
    ExportAllThread
    """
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()

    def __init__(self, circuit: MultiCircuit, drivers_list: List[DRIVER_OBJECTS], file_name: str):
        """
        Constructor
        :param circuit: Grid circuit
        :param drivers_list: list of VeraGrid simulation drivers
        :param file_name: name of the file where to save (.zip)
        """
        QThread.__init__(self)

        self.circuit: MultiCircuit = circuit

        self.drivers_list: List[DRIVER_OBJECTS] = drivers_list

        self.file_name: str = file_name

        self.valid: bool = False

        self.logger = Logger()

        self.error_msg: str = ""

        self.__cancel__ = False

    def run(self) -> None:
        """
        run the file save procedure
        """

        # try:
        export_drivers(drivers_list=self.drivers_list,
                       file_name=self.file_name,
                       text_func=self.progress_text.emit,
                       progress_func=self.progress_signal.emit,
                       logger=self.logger)

        self.valid = True

        # post events
        self.progress_text.emit('Done!')

        self.done_signal.emit()

    def cancel(self):
        """
        Cancel progress
        """
        self.__cancel__ = True



# MIT License
#
# Copyright (c) 2018 Ross Wilson
# Copyright (c) 2024, Santiago Peñate Vera
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
A tile source that serves OpenStreetMap tiles from server(s).
"""
from __future__ import  annotations
import math
from typing import Tuple, List
from VeraGrid.Gui.Diagrams.MapWidget.Tiles.tiles import Tiles


class CartoDbTiles(Tiles):
    """
    An object to source server tiles for pySlipQt.
    """

    def __init__(self, tiles_dir='open_street_map_tiles',
                 http_proxy=None,
                 tile_servers: List[str] | None = None,
                 url_path='/{Z}/{X}/{Y}.png',
                 name: str = 'Carto positron',
                 tile_set_short_name='CartoDb Dark Matter',
                 tile_set_version='1.0',
                 attribution="© CARTO, © OpenStreetMap contributors",
                 min_zoom: int = 0,
                 max_zoom: int = 22,
                 tile_width=256,
                 tile_height=256):
        """
        Override the base class for these tiles.
        Basically, just fill in the BaseTiles class with values from above
        and provide the Geo2Tile() and Tile2Geo() methods.
        :param tiles_dir: Tiles directory
        :param http_proxy: Proxy URL
        :param tile_servers: List of tile servers
        :param url_path: extra path to query the tiles
        :param name: Name of the tile set
        :param tile_set_short_name: Short name of the tile set
        :param tile_set_version: Version of the tile set
        :param attribution: Attribution string of the tile set
        :param min_zoom: Minimal zoom
        :param max_zoom: Maximum zoom (+1)
        :param tile_width: Width of the tile (px)
        :param tile_height: Height of the tile (px)
        """

        super().__init__(tile_set_name=name,
                         tile_set_short_name=tile_set_short_name,
                         tile_set_version=tile_set_version,
                         levels=[z for z in range(min_zoom, max_zoom)],
                         tile_width=tile_width,
                         tile_height=tile_height,
                         tiles_dir=tiles_dir,
                         servers=["https://basemaps.cartocdn.com/dark_all/"] if tile_servers is None else tile_servers,
                         url_path=url_path,
                         max_server_requests=2,
                         max_lru=10000,
                         http_proxy=http_proxy,
                         attribution=attribution)

        # get tile information into instance
        self.level = min(self.levels)
        self.num_tiles_x, self.num_tiles_y, self.ppd_x, self.ppd_y = self.GetInfo(self.level)

    def copy(self) -> "CartoDbTiles":
        """
        Copy of this object
        :return: CartoDbTiles
        """
        cpy = CartoDbTiles(tiles_dir=self.tiles_dir,
                           http_proxy=self.http_proxy,
                           tile_servers=self.servers.copy(),
                           url_path=self.url_path,
                           name=self.tile_set_name,
                           tile_set_short_name=self.tile_set_short_name,
                           tile_set_version=self.tile_set_version,
                           attribution=self.attribution_string,
                           min_zoom=self.min_level,
                           max_zoom=self.max_level,
                           tile_width=self.tile_width,
                           tile_height=self.tile_height
                           )

        return cpy

    def Geo2Tile(self, longitude: float, latitude: float) -> Tuple[float, float]:
        """
        Convert geo to tile fractional coordinates for level in use.
        geo  tuple of geo coordinates (longitude, latitude)
        Note that we assume the point *is* on the map!
        Code taken from [https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames]

        Lon./lat. to tile numbers:
        n = 2 ^ zoom
        x_tile = n * ((lon_deg + 180) / 360)
        y_tile = n * (1 - (log(tan(lat_rad) + sec(lat_rad)) / π)) / 2

        :return: This code returns the coordinate of the upper left point of the tile. (x_tile, y_tile)
        """
        lat_rad = math.radians(latitude)
        n = math.pow(2.0, self.level)
        x_tile = (longitude + 180.0) / 360.0 * n
        y_tile = ((1.0 - math.log(math.tan(lat_rad) + (1.0 / math.cos(lat_rad))) / math.pi) / 2.0) * n

        return x_tile, y_tile

    def Tile2Geo(self, x_tile: float, y_tile: float) -> Tuple[float, float]:
        """
        Convert tile fractional coordinates to geo for level in use.
        tile  a tuple (x_tile, y_tile) of tile fractional coordinates
        Note that we assume the point *is* on the map!
        Code taken from [https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames]

        n = 2 ^ zoom
        lon_deg = x_tile / n * 360.0 - 180.0
        lat_rad = arctan(sinh(π * (1 - 2 * y_tile / n)))
        lat_deg = lat_rad * 180.0 / π

        This returns the NW-corner of the square.
        Use the function with x_tile+1 and/or y_tile+1 to get the other corners.
        With x_tile+0.5 & y_tile+0.5 it will return the center of the tile.

        :return: This code returns the coordinate of the upper left point of the tile. (longitude, latitude)
        """

        n = 2.0 ** self.level
        longitude = x_tile / n * 360.0 - 180.0
        y_rad = math.atan(math.sinh(math.pi * (1 - 2 * y_tile / n)))
        latitude = math.degrees(y_rad)

        return longitude, latitude



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'block_editor.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QGraphicsView, QHBoxLayout,
    QHeaderView, QLabel, QMainWindow, QMenuBar,
    QPushButton, QSizePolicy, QSpacerItem, QSplitter,
    QTabWidget, QTableView, QToolBar, QTreeView,
    QVBoxLayout, QWidget)
from VeraGrid.Gui.Icons.icons_rc import *

class Ui_BlockEditorWindow(object):
    def setupUi(self, BlockEditorWindow):
        if not BlockEditorWindow.objectName():
            BlockEditorWindow.setObjectName(u"BlockEditorWindow")
        BlockEditorWindow.resize(1007, 510)
        self.block_editor_actionCheckModel = QAction(BlockEditorWindow)
        self.block_editor_actionCheckModel.setObjectName(u"block_editor_actionCheckModel")
        icon = QIcon()
        icon.addFile(u":/Icons/icons/calculator.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.block_editor_actionCheckModel.setIcon(icon)
        self.block_editor_actionCheckModel.setMenuRole(QAction.MenuRole.NoRole)
        self.centralwidget = QWidget(BlockEditorWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(6, 0, 6, 0)
        self.frame_3 = QFrame(self.centralwidget)
        self.frame_3.setObjectName(u"frame_3")
        self.frame_3.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_3.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_6 = QVBoxLayout(self.frame_3)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.verticalLayout_6.setContentsMargins(0, 6, 0, 0)
        self.splitter = QSplitter(self.frame_3)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Horizontal)
        self.frame_6 = QFrame(self.splitter)
        self.frame_6.setObjectName(u"frame_6")
        self.frame_6.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_6.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_4 = QVBoxLayout(self.frame_6)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(0, 0, 6, 0)
        self.tabWidget = QTabWidget(self.frame_6)
        self.tabWidget.setObjectName(u"tabWidget")
        self.tabWidget.setTabPosition(QTabWidget.TabPosition.North)
        self.tabWidget.setDocumentMode(True)
        self.tab = QWidget()
        self.tab.setObjectName(u"tab")
        self.verticalLayout_5 = QVBoxLayout(self.tab)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.libraryTreeView = QTreeView(self.tab)
        self.libraryTreeView.setObjectName(u"libraryTreeView")
        self.libraryTreeView.setFrameShape(QFrame.Shape.NoFrame)

        self.verticalLayout_5.addWidget(self.libraryTreeView)

        self.tabWidget.addTab(self.tab, "")
        self.tab_2 = QWidget()
        self.tab_2.setObjectName(u"tab_2")
        self.verticalLayout_2 = QVBoxLayout(self.tab_2)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.parametersTableView = QTableView(self.tab_2)
        self.parametersTableView.setObjectName(u"parametersTableView")
        self.parametersTableView.setFrameShape(QFrame.Shape.NoFrame)

        self.verticalLayout_2.addWidget(self.parametersTableView)

        self.tabWidget.addTab(self.tab_2, "")

        self.verticalLayout_4.addWidget(self.tabWidget)

        self.splitter.addWidget(self.frame_6)
        self.frame_7 = QFrame(self.splitter)
        self.frame_7.setObjectName(u"frame_7")
        self.frame_7.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_7.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_3 = QVBoxLayout(self.frame_7)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(6, 0, 0, 0)
        self.graphicsView = QGraphicsView(self.frame_7)
        self.graphicsView.setObjectName(u"graphicsView")

        self.verticalLayout_3.addWidget(self.graphicsView)

        self.splitter.addWidget(self.frame_7)

        self.verticalLayout_6.addWidget(self.splitter)


        self.verticalLayout.addWidget(self.frame_3)

        self.frame_2 = QFrame(self.centralwidget)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setMaximumSize(QSize(16777215, 40))
        self.frame_2.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout_2 = QHBoxLayout(self.frame_2)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.currently_editing_object_label = QLabel(self.frame_2)
        self.currently_editing_object_label.setObjectName(u"currently_editing_object_label")

        self.horizontalLayout_2.addWidget(self.currently_editing_object_label)

        self.horizontalSpacer_3 = QSpacerItem(859, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_3)

        self.doItButton = QPushButton(self.frame_2)
        self.doItButton.setObjectName(u"doItButton")

        self.horizontalLayout_2.addWidget(self.doItButton)


        self.verticalLayout.addWidget(self.frame_2)

        BlockEditorWindow.setCentralWidget(self.centralwidget)
        self.menuBar = QMenuBar(BlockEditorWindow)
        self.menuBar.setObjectName(u"menuBar")
        self.menuBar.setGeometry(QRect(0, 0, 1007, 23))
        BlockEditorWindow.setMenuBar(self.menuBar)
        self.toolBar = QToolBar(BlockEditorWindow)
        self.toolBar.setObjectName(u"toolBar")
        self.toolBar.setMovable(False)
        self.toolBar.setFloatable(False)
        BlockEditorWindow.addToolBar(Qt.ToolBarArea.TopToolBarArea, self.toolBar)

        self.toolBar.addAction(self.block_editor_actionCheckModel)

        self.retranslateUi(BlockEditorWindow)

        self.tabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(BlockEditorWindow)
    # setupUi

    def retranslateUi(self, BlockEditorWindow):
        BlockEditorWindow.setWindowTitle(QCoreApplication.translate("BlockEditorWindow", u"BlockEditorWindow", None))
        self.block_editor_actionCheckModel.setText(QCoreApplication.translate("BlockEditorWindow", u"CheckModel", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), QCoreApplication.translate("BlockEditorWindow", u"Library", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), QCoreApplication.translate("BlockEditorWindow", u"Parameters", None))
        self.currently_editing_object_label.setText(QCoreApplication.translate("BlockEditorWindow", u"Device:", None))
        self.doItButton.setText(QCoreApplication.translate("BlockEditorWindow", u"Do it!", None))
        self.toolBar.setWindowTitle(QCoreApplication.translate("BlockEditorWindow", u"toolBar", None))
    # retranslateUi

